# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from random import shuffle, random

from lollypop.define import Repeat, App
from lollypop.objects_track import Track
from lollypop.objects_album import Album
from lollypop.list import LinkedList
from lollypop.utils import emit_signal, get_default_storage_type
from lollypop.logger import Logger


class ShufflePlayer:
    """
        Shuffle player
        Manage shuffle tracks and party mode
    """

    def __init__(self):
        """
            Init shuffle player
        """
        # Albums to play, different from _albums for a better shuffle
        self.__to_play_albums = []
        # Albums who never have been played
        self.__not_played_albums = []
        # Tracks already played
        self.__history = []
        # Tracks already played by albums
        self.__already_played_tracks = {}
        # Party mode
        self._is_party = False
        App().settings.connect("changed::shuffle", self.__set_shuffle)
        self.connect("playback-added", self.__on_playback_added)
        self.connect("playback-setted", self.__on_playback_setted)
        self.connect("playback-removed", self.__on_playback_removed)

    def next(self):
        """
            Next shuffle track
            @return Track
        """
        if self.shuffle_has_next:
            track = self.__history.next.value
        elif self._albums:
            track = self.__get_next()
        else:
            track = Track()
        return track

    def prev(self):
        """
            Prev track based on history
            @return Track
        """
        if self.shuffle_has_prev:
            track = self.__history.prev.value
        else:
            track = self._current_track
        return track

    def set_party(self, party):
        """
            Set party mode on if party is True
            Play a new random track if not already playing
            @param party as bool
        """
        def start_party(*ignore):
            if self._albums:
                # Start a new song if not playing
                if self._current_track.id is None:
                    track = self.__get_tracks_random()
                    self.load(track)
                elif not self.is_playing:
                    self.play()
                else:
                    self.set_next()
            emit_signal(self, "loading-changed", False, Track())

        if party == self._is_party:
            return
        self._is_party = party

        if party:
            App().task_helper.run(self.set_party_ids, callback=(start_party,))
        else:
            # We want current album to continue playback
            self._albums = [self._current_track.album]
            emit_signal(self, "playback-setted", [])
            emit_signal(self, "playback-added",
                        self._current_track.album)
        if self._current_track.id is not None:
            self.set_next()
            self.set_prev()

    def set_party_ids(self):
        """
            Set party mode ids
        """
        if not self._is_party:
            return
        party_ids = App().settings.get_value("party-ids")
        storage_type = get_default_storage_type()
        album_ids = App().albums.get_ids(party_ids, [], storage_type, False)
        emit_signal(self, "playback-setted", [])
        self._albums = []
        if album_ids:
            emit_signal(self, "loading-changed", True, Track())
        for album_id in album_ids:
            album = Album(album_id, [], [], False)
            self._albums.append(album)
        emit_signal(self, "playback-setted", list(self._albums))

    @property
    def is_party(self):
        """
            True if party mode on
            @return bool
        """
        return self._is_party

    @property
    def shuffle_has_next(self):
        """
            True if history provide a next track
            @return bool
        """
        return self.__history and self.__history.has_next

    @property
    def shuffle_has_prev(self):
        """
            True if history provide a prev track
            @return bool
        """
        return self.__history and self.__history.has_prev

#######################
# PROTECTED           #
#######################
    def _on_stream_start(self, bus, message):
        """
            On stream start add to shuffle history
        """
        if self._current_track.id is None or\
                self._current_track.id < 0:
            return
        # Add track to shuffle history if needed
        if App().settings.get_value("shuffle") or self._is_party:
            self.__add_to_shuffle_history(self._current_track)
            if self.__history:
                next = self.__history.next
                prev = self.__history.prev
                # Remove next track
                if next is not None and\
                        self._current_track == next.value:
                    next = self.__history.next
                    next.set_prev(self.__history)
                    self.__history = next
                # Remove previous track
                elif prev is not None and\
                        self._current_track == prev.value:
                    prev = self.__history.prev
                    prev.set_next(self.__history)
                    self.__history = prev
                # Add a new track
                elif self.__history.value != self._current_track:
                    new_list = LinkedList(self._current_track,
                                          None,
                                          self.__history)
                    self.__history = new_list
            else:
                # Initial history
                new_list = LinkedList(self._current_track)
                self.__history = new_list

#######################
# PRIVATE             #
#######################
    def __set_shuffle(self, settings, value):
        """
            Update next track
            @param settings as Gio.Settings
            @param value as GLib.Variant
        """
        self.__on_playback_setted(self, self.albums)
        if self._current_track.id is not None:
            self.set_next()

    def __get_next(self):
        """
            Next track in shuffle mode
            @return track as Track
        """
        try:
            if App().settings.get_value("shuffle") or self._is_party:
                if self._albums:
                    track = self.__get_tracks_random()
                    # All tracks done
                    # Try to get another one track after reseting history
                    if track.id is None:
                        self.__to_play_albums = list(self._albums)
                        shuffle(self.__to_play_albums)
                        repeat = App().settings.get_enum("repeat")
                        # Do not reset history if a new album is going to
                        # be added
                        if repeat not in [Repeat.AUTO_SIMILAR,
                                          Repeat.AUTO_RANDOM]:
                            self.__history = []
                            self.__already_played_tracks = {}
                        if repeat == Repeat.ALL:
                            return self.__get_next()
                    return track
        except Exception as e:
            Logger.error("ShufflePLayer::__get_next(): %s", e)
        return Track()

    def __get_tracks_random(self):
        """
            Return a random track and make sure it has never been played
            @return Track
        """
        # True if all albums have been played on time
        if not self.__not_played_albums:
            self.__not_played_albums = list(self.__to_play_albums)
        while self.__not_played_albums:
            album = self.__not_played_albums.pop(0)
            if not album.tracks:
                continue
            for track in sorted(album.tracks,
                                key=lambda *args: random()):
                if not self.__in_shuffle_history(track):
                    return track
            self.__to_play_albums.remove(album)
        if self.__to_play_albums:
            return self.__get_tracks_random()
        else:
            return Track()

    def __in_shuffle_history(self, track):
        """
            True if track in shuffle history
            @param track as Track
            @return bool
        """
        return track.album.id in self.__already_played_tracks.keys() and\
            track.id in self.__already_played_tracks[track.album.id]

    def __add_to_shuffle_history(self, track):
        """
            Add a track to shuffle history
            @param track as Track
        """
        if track.album.id not in self.__already_played_tracks.keys():
            self.__already_played_tracks[track.album.id] = []
        if track not in self.__already_played_tracks[track.album.id]:
            self.__already_played_tracks[track.album.id].append(track.id)

    def __on_playback_added(self, player, album):
        """
            Update shuffle for album
            @param player as Player
            @param album as Album
        """
        if App().settings.get_value("shuffle") or self._is_party:
            if album not in self.__to_play_albums:
                self.__to_play_albums.append(album)
                shuffle(self.__to_play_albums)
            if album not in self.__not_played_albums:
                self.__not_played_albums.append(album)
            # If album already playing or
            # if current track was last one
            if App().player.current_track.album == album or\
                    not self.__already_played_tracks:
                self.__add_to_shuffle_history(App().player.current_track)

    def __on_playback_setted(self, player, albums):
        """
            Update shuffle for album
            @param player as Player
            @param albums as [Album]
        """
        if App().settings.get_value("shuffle") or self._is_party:
            self.__to_play_albums = albums
            if albums:
                shuffle(self.__to_play_albums)
            self.__not_played_albums = []
            self.__already_played_tracks = {}
            if App().player.current_track.album in albums:
                self.__add_to_shuffle_history(App().player.current_track)

    def __on_playback_removed(self, player, album):
        """
            Update shuffle for album
            @param player as Player
            @param album as Album
        """
        if App().settings.get_value("shuffle") or self._is_party:
            if album in self.__to_play_albums:
                self.__to_play_albums.remove(album)
            if album in self.__not_played_albums:
                self.__not_played_albums.remove(album)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk, Pango

from lollypop.define import App
from lollypop.helper_signals import SignalsHelper, signals
from lollypop.utils import on_query_tooltip


class LabelPlayerWidget(Gtk.Label, SignalsHelper):
    """
        Gtk.Label auto updated with current player status
    """

    @signals
    def __init__(self, fullscreen=False, font_size=None):
        """
            Init label
            @param fullscreen as bool
        """
        Gtk.Label.__init__(self)
        self.__fullscreen = fullscreen
        self.__font_size = font_size
        self.set_ellipsize(Pango.EllipsizeMode.END)
        self.set_has_tooltip(True)
        self.connect("query-tooltip", on_query_tooltip)
        return [
            (App().player, "current-changed", "_on_current_changed"),
        ]

    def update(self):
        """
            Update label
        """
        # Some tags contain \n inside them
        # Filter this
        artists = ", ".join(
            App().player.current_track.artists).replace("\n", "")
        title = App().player.current_track.title.replace("\n", "")
        artist_text = GLib.markup_escape_text(artists)
        title_text = GLib.markup_escape_text(title)
        if self.__font_size is None:
            markup = "<b>%s</b>" % artist_text
            markup += "\n<span size='small' alpha='50000'>%s</span>" %\
                title_text
        else:
            markup = "<span font='%s' size='x-large' ><b>%s</b></span>" % (
                                                           self.__font_size,
                                                           artist_text)
            markup += "\n<span font='%s' size='x-large'>%s</span>"\
                % (self.__font_size, title_text)
        if self.__fullscreen:
            album_text = "%s" % GLib.markup_escape_text(
                App().player.current_track.album.name)
            if App().player.current_track.year:
                album_text += " (%s)" % App().player.current_track.year
            if self.__font_size is None:
                markup += "\n<span size='x-small' alpha='40000'>%s</span>" %\
                    album_text
            else:
                markup += "\n<span font='%s' alpha='40000'>%s</span>" %\
                    (self.__font_size, album_text)
        self.set_markup(markup)

#######################
# PROTECTED           #
#######################
    def _on_current_changed(self, player):
        """
            Update artwork and labels
            @param player as Player
        """
        self.update()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from random import shuffle
from gettext import gettext as _
from time import time

from lollypop.view_flowbox import FlowBoxView
from lollypop.define import App, Type, ViewType, OrderBy, ScanUpdate
from lollypop.widgets_artist_rounded import RoundedArtistWidget
from lollypop.objects_album import Album
from lollypop.utils import get_icon_name
from lollypop.helper_signals import SignalsHelper, signals_map


class RoundedArtistsView(FlowBoxView, SignalsHelper):
    """
        Show artists in a FlowBox
    """

    @signals_map
    def __init__(self, storage_type, view_type):
        """
            Init artist view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        FlowBoxView.__init__(self, storage_type, view_type)
        self.__time = time() + 10
        self.connect("destroy", self.__on_destroy)
        self._empty_icon_name = get_icon_name(Type.ARTISTS)
        return [
            (App().artist_art, "artist-artwork-changed",
             "_on_artist_artwork_changed"),
            (App().scanner, "updated", "_on_collection_updated")
        ]

    def populate(self, artist_ids=[]):
        """
            Populate view with artist id
            Show all artists if empty
            @param artist_ids as [int]
        """
        def on_load(artist_ids):
            FlowBoxView.populate(self, artist_ids)

        def load():
            return App().artists.get([], self.storage_type)

        if artist_ids:
            FlowBoxView.populate(self, artist_ids)
        else:
            App().task_helper.run(load, callback=(on_load,))

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"storage_type": self.storage_type,
                "view_type": self.view_type}

#######################
# PROTECTED           #
#######################
    def _get_child(self, item):
        """
            Get a child for view
            @param item as (int, str, str)
            @return row as SelectionListRow
        """
        if self.destroyed:
            return None
        widget = RoundedArtistWidget(item, self.view_type, self.font_height)
        self._box.insert(widget, -1)
        widget.show()
        return widget

    def _get_menu_widget(self, child):
        """
            Get menu widget
            @param child as AlbumSimpleWidget
            @return Gtk.Widget
        """
        from lollypop.widgets_menu import MenuBuilder
        from lollypop.menu_artist import ArtistMenu
        from lollypop.menu_similars import SimilarsMenu
        menu = ArtistMenu(child.data, self.storage_type, self.view_type,
                          App().window.folded)
        menu_widget = MenuBuilder(menu, False)
        menu_widget.show()
        menu_ext = SimilarsMenu(child.data)
        menu_ext.show()
        menu_widget.add_widget(menu_ext)
        return menu_widget

    def _on_child_activated(self, flowbox, child):
        """
            Navigate into child
            @param flowbox as Gtk.FlowBox
            @param child as Gtk.FlowBoxChild
        """
        App().window.container.show_view([Type.ARTISTS], [child.data],
                                         self.storage_type)

    def _on_tertiary_press_gesture(self, x, y, event):
        """
            Play artist
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        child = self._box.get_child_at_pos(x, y)
        if child is None or child.artwork is None:
            return
        album_ids = App().albums.get_ids([], [child.data],
                                         self.storage_type, False)
        albums = [Album(album_id) for album_id in album_ids]
        if albums:
            App().player.play_album_for_albums(albums[0], albums)

    def _on_artist_artwork_changed(self, art, prefix):
        """
            Update artwork if needed
            @param art as Art
            @param prefix as str
        """
        for child in self._box.get_children():
            if child.name == prefix:
                child.set_artwork()

    def _on_collection_updated(self, scanner, item, scan_update):
        """
            Add/remove artist to/from list
            @param scanner as CollectionScanner
            @param item as CollectionItem
            @param scan_update as ScanUpdate
        """
        # On first update, ignore notifications for 10 seconds
        # Next, ignore notifications for 120 seconds
        if scan_update == ScanUpdate.ADDED and time() > self.__time:
            self.__time = time() + 120
            App().window.container.show_notification(
                    _("New artists available"),
                    [_("Refresh")],
                    [App().window.container.reload_view])
        elif scan_update == ScanUpdate.REMOVED:
            for child in self._box.get_children():
                if child.data in item.new_album_artist_ids:
                    child.destroy()
                    break

#######################
# PRIVATE             #
#######################
    def __on_destroy(self, widget):
        """
            Stop loading
            @param widget as Gtk.Widget
        """
        RoundedArtistsView.stop(self)


class RoundedArtistsViewWithBanner(RoundedArtistsView):
    """
        Show rounded artist view with a banner
    """

    def __init__(self, storage_type, view_type):
        """
            Init artist view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        from lollypop.widgets_banner_flowbox import FlowboxBannerWidget
        RoundedArtistsView.__init__(self, storage_type,
                                    view_type | ViewType.OVERLAY)
        self.__banner = FlowboxBannerWidget([Type.ARTISTS], [], self.view_type)
        self.__banner.show()
        self.__banner.connect("play-all", self.__on_banner_play_all)
        self.add_widget(self._box, self.__banner)

#######################
# PROTECTED           #
#######################
    def _on_map(self, widget):
        """
            Set initial view state
            @param widget as GtK.Widget
        """
        RoundedArtistsView._on_map(self, widget)
        if self.view_type & ViewType.SCROLLED:
            self.scrolled.grab_focus()

#######################
# PRIVATE             #
#######################
    def __on_banner_play_all(self, banner, random):
        """
            Play all albums
            @param banner as AlbumsBannerWidget
            @param random as bool
        """
        album_ids = App().albums.get_ids([], [], self.storage_type,
                                         False, OrderBy.ARTIST_YEAR)
        if not album_ids:
            return
        albums = [Album(album_id) for album_id in album_ids]
        if random:
            shuffle(albums)
            App().player.play_album_for_albums(albums[0], albums)
        else:
            App().player.play_album_for_albums(albums[0], albums)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json

from lollypop.helper_web_base import BaseWebHelper
from lollypop.define import App
from lollypop.utils import emit_signal
from lollypop.logger import Logger


class InvidiousWebHelper(BaseWebHelper):
    """
        Invidious helper
    """

    __BAD_SCORE = 1000000
    __SEARCH = "api/v1/search?q=%s"
    __VIDEO = "api/v1/videos/%s"

    def __init__(self):
        """
            Init heApper
        """
        BaseWebHelper.__init__(self)
        self.__server = App().settings.get_value(
            "invidious-server").get_string().strip("/")

    def get_uri_content(self, uri, cancellable):
        """
            Get content uri
            @param uri as str
            @param cancellable as Gio.Cancellable
            @return content uri as str/None
        """
        Logger.info("Loading %s with Invidious %s", uri, self.__server)
        youtube_id = uri.replace("https://www.youtube.com/watch?v=", "")
        video = self.__VIDEO % youtube_id
        uri = "%s/%s" % (self.__server, video)
        App().task_helper.load_uri_content(uri,
                                           cancellable,
                                           self.__on_uri_content)

#######################
# PRIVATE             #
#######################
    def __on_uri_content(self, uri, status, content):
        """
            Emit signal for content
            @param uri as str
            @param status as bool
            @param content as bytes
        """
        try:
            content_uri = ""
            if status:
                decode = json.loads(content.decode("utf-8"))
                for item in decode["adaptiveFormats"]:
                    if item["container"] == "webm" and\
                            item["encoding"] == "opus":
                        content_uri = item["url"]
                        break
        except Exception as e:
            Logger.warning("InvidiousWebHelper::__on_uri_content(): %s", e)
        emit_signal(self, "uri-content-loaded", content_uri)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (c) 2016 Gaurav Narula
# Copyright (c) 2016 Felipe Borges <felipeborges@gnome.org>
# Copyright (c) 2013 Arnel A. Borja <kyoushuu@yahoo.com>
# Copyright (c) 2013 Vadim Rutkovsky <vrutkovs@redhat.com>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, Gst, GLib, Gtk

from random import randint

from lollypop.logger import Logger
from lollypop.define import App, ArtSize, Repeat, Notifications
from lollypop.objects_track import Track


class Server:

    def __init__(self, con, path):
        method_outargs = {}
        method_inargs = {}
        for interface in Gio.DBusNodeInfo.new_for_xml(self.__doc__).interfaces:

            for method in interface.methods:
                method_outargs[method.name] = "(" + "".join(
                              [arg.signature for arg in method.out_args]) + ")"
                method_inargs[method.name] = tuple(
                    arg.signature for arg in method.in_args)

            con.register_object(object_path=path,
                                interface_info=interface,
                                method_call_closure=self.on_method_call)

        self.method_inargs = method_inargs
        self.method_outargs = method_outargs

    def on_method_call(self,
                       connection,
                       sender,
                       object_path,
                       interface_name,
                       method_name,
                       parameters,
                       invocation):

        args = list(parameters.unpack())
        for i, sig in enumerate(self.method_inargs[method_name]):
            if sig == "h":
                msg = invocation.get_message()
                fd_list = msg.get_unix_fd_list()
                args[i] = fd_list.get(args[i])

        try:
            result = getattr(self, method_name)(*args)

            # out_args is atleast (signature1).
            # We therefore always wrap the result as a tuple.
            # Refer to https://bugzilla.gnome.org/show_bug.cgi?id=765603
            result = (result,)

            out_args = self.method_outargs[method_name]
            if out_args != "()":
                variant = GLib.Variant(out_args, result)
                invocation.return_value(variant)
            else:
                invocation.return_value(None)
        except:
            pass


class MPRIS(Server):
    """
    <!DOCTYPE node PUBLIC
    "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
    "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
    <node>
        <interface name="org.freedesktop.DBus.Introspectable">
            <method name="Introspect">
                <arg name="data" direction="out" type="s"/>
            </method>
        </interface>
        <interface name="org.freedesktop.DBus.Properties">
            <method name="Get">
                <arg name="interface" direction="in" type="s"/>
                <arg name="property" direction="in" type="s"/>
                <arg name="value" direction="out" type="v"/>
            </method>
            <method name="Set">
                <arg name="interface_name" direction="in" type="s"/>
                <arg name="property_name" direction="in" type="s"/>
                <arg name="value" direction="in" type="v"/>
            </method>
            <method name="GetAll">
                <arg name="interface" direction="in" type="s"/>
                <arg name="properties" direction="out" type="a{sv}"/>
            </method>
        </interface>
        <interface name="org.mpris.MediaPlayer2">
            <method name="Raise">
            </method>
            <method name="Quit">
            </method>
            <property name="CanQuit" type="b" access="read" />
            <property name="Fullscreen" type="b" access="readwrite" />
            <property name="CanRaise" type="b" access="read" />
            <property name="HasTrackList" type="b" access="read"/>
            <property name="Identity" type="s" access="read"/>
            <property name="DesktopEntry" type="s" access="read"/>
            <property name="SupportedUriSchemes" type="as" access="read"/>
            <property name="SupportedMimeTypes" type="as" access="read"/>
        </interface>
        <interface name="org.mpris.MediaPlayer2.Player">
            <method name="Next"/>
            <method name="Previous"/>
            <method name="Pause"/>
            <method name="PlayPause"/>
            <method name="Stop"/>
            <method name="Play"/>
            <method name="Seek">
                <arg direction="in" name="Offset" type="x"/>
            </method>
            <method name="SetPosition">
                <arg direction="in" name="TrackId" type="o"/>
                <arg direction="in" name="Position" type="x"/>
            </method>
            <method name="OpenUri">
                <arg direction="in" name="Uri" type="s"/>
            </method>
            <signal name="Seeked">
                <arg name="Position" type="x"/>
            </signal>
            <property name="PlaybackStatus" type="s" access="read"/>
            <property name="LoopStatus" type="s" access="readwrite"/>
            <property name="Rate" type="d" access="readwrite"/>
            <property name="Shuffle" type="b" access="readwrite"/>
            <property name="Metadata" type="a{sv}" access="read">
            </property>
            <property name="Volume" type="d" access="readwrite"/>
            <property name="Position" type="x" access="read"/>
            <property name="MinimumRate" type="d" access="read"/>
            <property name="MaximumRate" type="d" access="read"/>
            <property name="CanGoNext" type="b" access="read"/>
            <property name="CanGoPrevious" type="b" access="read"/>
            <property name="CanPlay" type="b" access="read"/>
            <property name="CanPause" type="b" access="read"/>
            <property name="CanSeek" type="b" access="read"/>
            <property name="CanControl" type="b" access="read"/>
        </interface>
        <interface name="org.mpris.MediaPlayer2.ExtensionSetRatings">
            <method name="SetRating">\
                <arg direction="in" name="TrackId" type="o"/>
                <arg direction="in" name="Rating" type="d"/>\
            </method>\
            <property name="HasRatingsExtension" type="b" access="read"/>\
        </interface>
    </node>
    """
    __MPRIS_IFACE = "org.mpris.MediaPlayer2"
    __MPRIS_PLAYER_IFACE = "org.mpris.MediaPlayer2.Player"
    __MPRIS_RATINGS_IFACE = "org.mpris.MediaPlayer2.ExtensionSetRatings"
    __MPRIS_LOLLYPOP = "org.mpris.MediaPlayer2.Lollypop"
    __MPRIS_PATH = "/org/mpris/MediaPlayer2"

    def __init__(self, app):
        self.__app = app
        self.__rating = None
        self.__lollypop_id = 0
        self.__metadata = {"mpris:trackid": GLib.Variant(
            "o",
            "/org/mpris/MediaPlayer2/TrackList/NoTrack")}
        self.__track_id = self.__get_media_id(0)
        self.__bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        Gio.bus_own_name_on_connection(self.__bus,
                                       self.__MPRIS_LOLLYPOP,
                                       Gio.BusNameOwnerFlags.NONE,
                                       None,
                                       None)
        Server.__init__(self, self.__bus, self.__MPRIS_PATH)
        App().player.connect("current-changed", self.__on_current_changed)
        App().player.connect("seeked", self.__on_seeked)
        App().player.connect("status-changed", self.__on_status_changed)
        App().player.connect("volume-changed", self.__on_volume_changed)
        App().player.connect("rate-changed", self.__on_rate_changed)
        App().settings.connect("changed::shuffle", self.__on_shuffle_changed)
        App().settings.connect("changed::repeat", self.__on_repeat_changed)

    def Raise(self):
        self.__app.window.present_with_time(Gtk.get_current_event_time())

    def Quit(self):
        self.__app.quit()

    def Next(self):
        App().player.next()
        if App().settings.get_enum("notifications") == Notifications.MPRIS:
            App().notify.send_track(App().player.current_track)

    def Previous(self):
        App().player.prev()
        if App().settings.get_enum("notifications") == Notifications.MPRIS:
            App().notify.send_track(App().player.current_track)

    def Pause(self):
        App().player.pause()

    def PlayPause(self):
        App().player.play_pause()

    def Stop(self):
        App().player.stop()

    def Play(self):
        if App().player.current_track.id is None:
            App().player.set_party(True)
        else:
            App().player.play()

    def SetPosition(self, track_id, position):
        App().player.seek(position / 1000)

    def OpenUri(self, uri):
        track_id = App().tracks.get_id_by_uri(uri)
        if track_id:
            App().player.load(Track(track_id))

    def Seek(self, offset):
        position = App().player.position
        App().player.seek(position + offset / 1000)

    def Seeked(self, position):
        self.__bus.emit_signal(
            None,
            self.__MPRIS_PATH,
            self.__MPRIS_PLAYER_IFACE,
            "Seeked",
            GLib.Variant.new_tuple(GLib.Variant("x", position)))

    def SetRating(self, track_id, rating):
        # We don't currently care about the trackId since
        # we have not yet implemented the TrackList interface.
        App().player.current_track.set_rate(int(rating * 5))

    def Get(self, interface, property_name):
        if property_name in ["CanQuit", "CanRaise", "CanSeek",
                             "CanControl", "HasRatingsExtension"]:
            return GLib.Variant("b", True)
        elif property_name == "HasTrackList":
            return GLib.Variant("b", False)
        elif property_name == "Shuffle":
            return App().settings.get_value("shuffle")
        elif property_name in ["Rate", "MinimumRate", "MaximumRate"]:
            return GLib.Variant("d", 1.0)
        elif property_name == "Identity":
            return GLib.Variant("s", "Lollypop")
        elif property_name == "DesktopEntry":
            return GLib.Variant("s", "org.gnome.Lollypop")
        elif property_name == "SupportedUriSchemes":
            return GLib.Variant("as", ["file", "http"])
        elif property_name == "SupportedMimeTypes":
            return GLib.Variant("as", ["application/ogg",
                                       "audio/x-vorbis+ogg",
                                       "audio/x-flac",
                                       "audio/mpeg"])
        elif property_name == "PlaybackStatus":
            return GLib.Variant("s", self.__get_status())
        elif property_name == "LoopStatus":
            repeat = App().settings.get_enum("repeat")
            if repeat == Repeat.ALL:
                value = "Playlist"
            elif repeat == Repeat.TRACK:
                value = "Track"
            else:
                value = "None"
            return GLib.Variant("s", value)
        elif property_name == "Metadata":
            return GLib.Variant("a{sv}", self.__metadata)
        elif property_name == "Volume":
            return GLib.Variant("d", App().player.volume)
        elif property_name == "Position":
            return GLib.Variant(
                "x",
                App().player.position * 1000)
        elif property_name in ["CanGoNext", "CanGoPrevious",
                               "CanPlay", "CanPause"]:
            return GLib.Variant("b", App().player.current_track.id is not None)

    def GetAll(self, interface):
        ret = {}
        if interface == self.__MPRIS_IFACE:
            for property_name in ["CanQuit",
                                  "CanRaise",
                                  "HasTrackList",
                                  "Identity",
                                  "DesktopEntry",
                                  "SupportedUriSchemes",
                                  "SupportedMimeTypes"]:
                ret[property_name] = self.Get(interface, property_name)
        elif interface == self.__MPRIS_PLAYER_IFACE:
            for property_name in ["PlaybackStatus",
                                  "LoopStatus",
                                  "Rate",
                                  "Shuffle",
                                  "Metadata",
                                  "Volume",
                                  "Position",
                                  "MinimumRate",
                                  "MaximumRate",
                                  "CanGoNext",
                                  "CanGoPrevious",
                                  "CanPlay",
                                  "CanPause",
                                  "CanSeek",
                                  "CanControl"]:
                ret[property_name] = self.Get(interface, property_name)
        elif interface == self.__MPRIS_RATINGS_IFACE:
            ret["HasRatingsExtension"] = GLib.Variant("b", True)
        return ret

    def Set(self, interface, property_name, new_value):
        if property_name == "Volume":
            App().player.set_volume(new_value)
        elif property_name == "Shuffle":
            App().settings.set_value("shuffle", GLib.Variant("b", new_value))
        elif property_name == "LoopStatus":
            if new_value == "Playlist":
                value = Repeat.ALL
            elif new_value == "Track":
                value = Repeat.TRACK
            else:
                value = Repeat.NONE
            App().settings.set_enum("repeat", value)

    def PropertiesChanged(self, interface_name, changed_properties,
                          invalidated_properties):
        self.__bus.emit_signal(None,
                               self.__MPRIS_PATH,
                               "org.freedesktop.DBus.Properties",
                               "PropertiesChanged",
                               GLib.Variant.new_tuple(
                                   GLib.Variant("s", interface_name),
                                   GLib.Variant("a{sv}", changed_properties),
                                   GLib.Variant("as", invalidated_properties)))

    def Introspect(self):
        return self.__doc__

#######################
# PRIVATE             #
#######################

    def __get_media_id(self, track_id):
        """
            TrackId's must be unique even up to
            the point that if you repeat a song
            it must have a different TrackId.
        """
        track_id = track_id + randint(10000000, 90000000)
        return GLib.Variant("o", "/org/gnome/Lollypop/TrackId/%s" % track_id)

    def __get_status(self):
        state = App().player.get_status()
        if state == Gst.State.PLAYING:
            return "Playing"
        elif state == Gst.State.PAUSED:
            return "Paused"
        else:
            return "Stopped"

    def __update_metadata(self):
        self.__metadata = {}
        if App().player.current_track.id is None or\
                self.__get_status() == "Stopped":
            self.__metadata = {"mpris:trackid": GLib.Variant(
                "o",
                "/org/mpris/MediaPlayer2/TrackList/NoTrack")}
        else:
            self.__metadata["mpris:trackid"] = self.__track_id
            track_number = App().player.current_track.number
            if track_number is None:
                track_number = 1
            self.__metadata["xesam:trackNumber"] = GLib.Variant(
                "i",
                track_number)
            self.__metadata["xesam:title"] = GLib.Variant(
                "s",
                App().player.current_track.name)
            self.__metadata["xesam:album"] = GLib.Variant(
                "s",
                App().player.current_track.album.name)
            self.__metadata["xesam:artist"] = GLib.Variant(
                "as",
                App().player.current_track.artists)
            self.__metadata["xesam:albumArtist"] = GLib.Variant(
                "as",
                App().player.current_track.album_artists)
            self.__metadata["mpris:length"] = GLib.Variant(
                "x",
                App().player.current_track.duration * 1000)
            self.__metadata["xesam:genre"] = GLib.Variant(
                "as",
                App().player.current_track.genres)
            self.__metadata["xesam:url"] = GLib.Variant(
                "s",
                App().player.current_track.uri)
            self.__metadata["xesam:userRating"] = GLib.Variant(
                "d",
                App().player.current_track.rate / 5)
            cover_path = App().album_art.get_cache_path(
                    App().player.current_track.album,
                    ArtSize.MPRIS, ArtSize.MPRIS)
            if cover_path is not None:
                self.__metadata["mpris:artUrl"] = GLib.Variant(
                    "s",
                    "file://" + cover_path)

    def __on_seeked(self, player, position):
        self.Seeked(position * 1000)

    def __on_volume_changed(self, player, data=None):
        self.PropertiesChanged(self.__MPRIS_PLAYER_IFACE,
                               {"Volume": GLib.Variant("d",
                                                       App().player.volume), },
                               [])

    def __on_shuffle_changed(self, settings, value):
        properties = {"Shuffle": App().settings.get_value("shuffle")}
        self.PropertiesChanged(self.__MPRIS_PLAYER_IFACE, properties, [])

    def __on_repeat_changed(self, settings, value):
        repeat = App().settings.get_enum("repeat")
        if repeat == Repeat.ALL:
            value = "Playlist"
        elif repeat == Repeat.TRACK:
            value = "Track"
        else:
            value = "None"
        properties = {"LoopStatus": GLib.Variant("s", value)}
        self.PropertiesChanged(self.__MPRIS_PLAYER_IFACE, properties, [])

    def __on_rate_changed(self, player, rated_track_id, rating):
        # We only care about the current Track's rating.
        if rated_track_id == self.__lollypop_id:
            self.__rating = rating
            self.__update_metadata()
            properties = {"Metadata": GLib.Variant("a{sv}", self.__metadata)}
            self.PropertiesChanged(self.__MPRIS_PLAYER_IFACE, properties, [])

    def __on_current_changed(self, player):
        if App().player.current_track.id is None:
            self.__lollypop_id = 0
        else:
            self.__lollypop_id = App().player.current_track.id
        # We only need to recalculate a new trackId at song changes.
        self.__track_id = self.__get_media_id(self.__lollypop_id)
        self.__rating = None
        self.__update_metadata()
        properties = {"Metadata": GLib.Variant("a{sv}", self.__metadata),
                      "CanPlay": GLib.Variant("b", True),
                      "CanPause": GLib.Variant("b", True),
                      "CanGoNext": GLib.Variant("b", True),
                      "CanGoPrevious": GLib.Variant("b", True)}
        try:
            self.PropertiesChanged(self.__MPRIS_PLAYER_IFACE, properties, [])
        except Exception as e:
            Logger.error("MPRIS::__on_current_changed(): %s" % e)

    def __on_status_changed(self, data=None):
        properties = {"PlaybackStatus": GLib.Variant("s", self.__get_status())}
        self.PropertiesChanged(self.__MPRIS_PLAYER_IFACE, properties, [])



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Pango


class AppNotification(Gtk.Revealer):
    """
        Show a notification to user with a button connected to an action
    """

    def __init__(self, message, button_labels, actions, timeout=None):
        """
            Init notification
            @param message as str
            @param button_label as [str]
            @param action as [callback]
            @param timeout as int
        """
        Gtk.Revealer.__init__(self)
        widget = Gtk.Grid()
        widget.get_style_context().add_class("app-notification")
        widget.set_column_spacing(5)
        label = Gtk.Label.new(message)
        label.set_line_wrap_mode(Pango.WrapMode.WORD)
        label.set_line_wrap(True)
        widget.add(label)
        for i in range(0, len(button_labels)):
            button = Gtk.Button.new()
            button.set_label(button_labels[i])
            button.connect("clicked", self.__on_button_clicked, actions[i])
            button.set_property("valign", Gtk.Align.START)
            widget.add(button)
        button = Gtk.Button.new_from_icon_name("window-close-symbolic",
                                               Gtk.IconSize.BUTTON)
        button.connect("clicked", self.__on_button_clicked, None)
        button.set_property("valign", Gtk.Align.START)
        widget.add(button)
        widget.show_all()
        self.add(widget)
        self.set_property("halign", Gtk.Align.CENTER)
        self.set_property("valign", Gtk.Align.START)
        if timeout is not None:
            GLib.timeout_add(timeout, self.destroy)

#######################
# PRIVATE             #
#######################
    def __on_button_clicked(self, button, action=None):
        """
            Execute action
            @param button as Gtk.Button
            @param action as callback
        """
        self.set_reveal_child(False)
        GLib.timeout_add(1000, self.destroy)
        if action is not None:
            action()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.logger import Logger
from lollypop.utils_file import decodeUnicode, splitUnicode
from lollypop.tag_frame import FrameTag
from lollypop.define import UTF_16_ENCODING, UTF_16BE_ENCODING


class FrameLangTag(FrameTag):
    """
        Bytes representing a text with lang frame
    """

    def __init__(self, bytes):
        """
            Init tag reader
            @param bytes as bytes
        """
        FrameTag.__init__(self, bytes)

    @property
    def string(self):
        """
            String representation of data
            @return str/None
        """
        try:
            (d, t) = splitUnicode(self.frame, self.encoding)
            if self.encoding in [UTF_16_ENCODING, UTF_16BE_ENCODING]:
                (start, end) = t.split(b"\x00\x00\xff", 1)
                end = b"\xff" + end
            else:
                (start, end) = t.split(b"\x00", 1)
            return decodeUnicode(end, self.encoding)
        except Exception as e:
            Logger.error("FrameLangTag::string: %s, %s", e, self.frame)
            return ""



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, GLib, Gio, GdkPixbuf, GObject

from gettext import gettext as _

from lollypop.logger import Logger
from lollypop.utils import emit_signal
from lollypop.define import App, ArtSize, ArtBehaviour


class ArtworkSearchChild(Gtk.FlowBoxChild):
    """
        Child for ArtworkSearch
    """

    __gsignals__ = {
        "hidden": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    def __init__(self, api, view_type):
        """
            Init child
            @param api as str
            @param view_type as ViewType
        """
        Gtk.FlowBoxChild.__init__(self)
        self.__bytes = None
        self.__view_type = view_type
        self.__api = api
        self.__image = Gtk.Image()
        self.__image.show()
        self.__label = Gtk.Label()
        self.__label.show()
        grid = Gtk.Grid()
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        grid.show()
        grid.add(self.__image)
        grid.add(self.__label)
        grid.set_row_spacing(5)
        self.__image.get_style_context().add_class("cover-frame")
        self.__image.set_property("halign", Gtk.Align.CENTER)
        self.__image.set_property("valign", Gtk.Align.CENTER)
        self.add(grid)

    def populate(self, bytes, art_manager, art_size):
        """
            Populate images with bytes
            @param bytes as bytes
            @param art_manager as ArtworkManager
            @param art_size as int
            @return bool if success
        """
        try:
            scale_factor = self.get_scale_factor()
            gbytes = GLib.Bytes.new(bytes)
            stream = Gio.MemoryInputStream.new_from_bytes(gbytes)
            if stream is not None:
                pixbuf = GdkPixbuf.Pixbuf.new_from_stream(stream, None)
                if self.__api is None:
                    text = "%sx%s" % (pixbuf.get_width(),
                                      pixbuf.get_height())
                else:
                    text = "%s: %sx%s" % (self.__api,
                                          pixbuf.get_width(),
                                          pixbuf.get_height())
                self.__label.set_text(text)
                pixbuf = art_manager.load_behaviour(
                                                pixbuf,
                                                art_size * scale_factor,
                                                art_size * scale_factor,
                                                ArtBehaviour.CROP)
                stream.close()
                self.__bytes = bytes
                surface = Gdk.cairo_surface_create_from_pixbuf(
                                                       pixbuf,
                                                       scale_factor,
                                                       None)
                self.__image.set_from_surface(surface)
                return True
        except Exception as e:
            Logger.error("ArtworkSearchChild::__get_image: %s" % e)
        return False

    @property
    def bytes(self):
        """
            Get bytes associated to widget
            @return bytes
        """
        return self.__bytes


class ArtworkSearchWidget(Gtk.Grid):
    """
        Search for artwork
    """

    def __init__(self, view_type):
        """
            Init widget
            @param view_type as ViewType
        """
        Gtk.Grid.__init__(self)
        self.__view_type = view_type
        self.__timeout_id = None
        self.__uri_artwork_id = None
        self._loaders = 0
        self._cancellable = Gio.Cancellable()
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/ArtworkSearch.ui")
        builder.connect_signals(self)
        widget = builder.get_object("widget")
        self.__stack = builder.get_object("stack")
        self.__entry = builder.get_object("entry")
        self.__spinner = builder.get_object("spinner")

        self._flowbox = Gtk.FlowBox()
        self._flowbox.set_selection_mode(Gtk.SelectionMode.SINGLE)
        self._flowbox.connect("child-activated", self._on_activate)
        self._flowbox.set_max_children_per_line(100)
        self._flowbox.set_property("row-spacing", 10)
        self._flowbox.set_property("valign", Gtk.Align.START)
        self._flowbox.set_vexpand(True)
        self._flowbox.set_homogeneous(True)
        self._flowbox.show()

        self.__label = builder.get_object("label")
        self.__label.set_text(_("Select artwork"))

        if App().window.folded:
            self._art_size = ArtSize.MEDIUM
            widget.add(self._flowbox)
        else:
            self._art_size = ArtSize.BANNER
            scrolled = Gtk.ScrolledWindow.new()
            scrolled.show()
            scrolled.set_size_request(700, 400)
            scrolled.set_vexpand(True)
            viewport = Gtk.Viewport.new()
            viewport.show()
            viewport.add(self._flowbox)
            scrolled.add(viewport)
            widget.add(scrolled)
        self.add(widget)
        self.connect("unmap", self.__on_unmap)

    def populate(self):
        """
            Populate view
        """
        try:
            grid = Gtk.Grid()
            grid.set_orientation(Gtk.Orientation.VERTICAL)
            grid.show()
            grid.set_row_spacing(5)
            image = Gtk.Image.new_from_icon_name("edit-clear-all-symbolic",
                                                 Gtk.IconSize.INVALID)
            image.set_pixel_size(self._art_size)
            context = image.get_style_context()
            context.add_class("cover-frame")
            padding = context.get_padding(Gtk.StateFlags.NORMAL)
            border = context.get_border(Gtk.StateFlags.NORMAL)
            image.set_size_request(self._art_size + padding.left +
                                   padding.right + border.left + border.right,
                                   self._art_size + padding.top +
                                   padding.bottom + border.top + border.bottom)
            image.show()
            label = Gtk.Label.new(_("Remove"))
            label.show()
            grid.add(image)
            grid.add(label)
            grid.set_property("valign", Gtk.Align.CENTER)
            grid.set_property("halign", Gtk.Align.CENTER)
            self._flowbox.add(grid)
            self._search_for_artwork()
        except Exception as e:
            Logger.error("ArtworkSearchWidget::populate(): %s", e)

    def stop(self):
        """
            Stop loading
        """
        self._cancellable.cancel()

    @property
    def view_type(self):
        """
            Get view type
            @return ViewType
        """
        return self.__view_type

#######################
# PROTECTED           #
#######################
    def _search_for_artwork(self):
        """
            Search artwork on the web
        """
        self._loaders = 0
        self._cancellable = Gio.Cancellable()
        self.__spinner.start()

    def _save_from_filename(self, filename):
        pass

    def _on_button_clicked(self, button):
        """
            Show file chooser
            @param button as Gtk.button
        """
        dialog = Gtk.FileChooserDialog()
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        dialog.add_buttons(Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        dialog.set_transient_for(App().window)
        file_filter = Gtk.FileFilter.new()
        file_filter.add_pixbuf_formats()
        dialog.set_filter(file_filter)
        emit_signal(self, "hidden", True)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            self._save_from_filename(dialog.get_filename())
        dialog.destroy()

    def _get_current_search(self):
        """
            Return current searches
            @return str
        """
        return self.__entry.get_text()

    def _on_search_changed(self, entry):
        """
            Launch search based on current text
            @param entry as Gtk.Entry
        """
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
        self.__timeout_id = GLib.timeout_add(1000,
                                             self.__on_search_timeout)

    def _on_activate(self, flowbox, child):
        pass

    def _on_uri_artwork_found(self, art, uris):
        """
            Load content in view
            @param art as Art
            @param uris as (str, str)/None
        """
        if uris:
            (uri, api) = uris.pop(0)
            App().task_helper.load_uri_content(uri,
                                               self._cancellable,
                                               self.__on_load_uri_content,
                                               api,
                                               uris,
                                               self._cancellable)
        else:
            self._loaders -= 1
            if self._loaders == 0:
                self.__spinner.stop()

#######################
# PRIVATE             #
#######################
    def __add_pixbuf(self, content, api):
        """
            Add content to view
            @param content as bytes
            @param api as str
        """
        child = ArtworkSearchChild(api, self.__view_type)
        child.show()
        status = child.populate(content, self.art, self._art_size)
        if status:
            child.set_name("web")
            self._flowbox.add(child)
        else:
            child.destroy()

    def __on_unmap(self, widget):
        """
            Cancel loading
            @param widget as Gtk.Widget
        """
        self._cancellable.cancel()

    def __on_load_uri_content(self, uri, loaded, content,
                              api, uris, cancellable):
        """
            Add loaded pixbuf
            @param uri as str
            @param loaded as bool
            @param content as bytes
            @param api as str
            @param uris as [str]
            @param cancellable as Gio.Cancellable
        """
        try:
            if loaded:
                self.__add_pixbuf(content, api)
            if uris and not cancellable.is_cancelled():
                (uri, api) = uris.pop(0)
                App().task_helper.load_uri_content(uri,
                                                   cancellable,
                                                   self.__on_load_uri_content,
                                                   api,
                                                   uris,
                                                   cancellable)
            else:
                self._loaders -= 1
        except Exception as e:
            self._loaders -= 1
            Logger.warning(
                "ArtworkSearchWidget::__on_load_uri_content(): %s", e)
        if self._loaders == 0:
            self.__spinner.stop()

    def __on_search_timeout(self):
        """
            Populate widget
        """
        self.__timeout_id = None
        self._cancellable.cancel()
        self._cancellable = Gio.Cancellable()
        for child in self._flowbox.get_children():
            if child.get_name() == "web":
                child.destroy()
        GLib.timeout_add(500, self._search_for_artwork)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GObject, Gtk

from lollypop.logger import Logger
from lollypop.utils import emit_signal
from lollypop.widgets_artwork import ArtworkSearchWidget, ArtworkSearchChild
from lollypop.define import App, ArtSize, StorageType
from lollypop.helper_signals import SignalsHelper, signals_map


class ArtistArtworkSearchWidget(ArtworkSearchWidget, SignalsHelper):
    """
        Search for artist artwork
    """

    __gsignals__ = {
        "hidden": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    @signals_map
    def __init__(self, artist_id, view_type, in_menu=False):
        """
            Init search
            @param artist_id as int
            @param view_type as ViewType
        """
        ArtworkSearchWidget.__init__(self, view_type)
        self.__artist = App().artists.get_name(artist_id)
        return [
            (App().artist_art, "uri-artwork-found", "_on_uri_artwork_found")
        ]

    @property
    def art(self):
        """
            Get related artwork
            @return ArtworkManager
        """
        return App().artist_art

#######################
# PROTECTED           #
#######################
    def _save_from_filename(self, filename):
        """
            Save filename as album artwork
            @param button as Gtk.button
        """
        try:
            f = Gio.File.new_for_path(filename)
            (status, data, tag) = f.load_contents()
            if status:
                App().artist_art.add(self.__artist, data,
                                     StorageType.COLLECTION)
        except Exception as e:
            Logger.error(
                "ArtistArtworkSearchWidget::_save_from_filename(): %s" % e)

    def _get_current_search(self):
        """
            Return current searches
            @return str
        """
        search = ArtworkSearchWidget._get_current_search(self)
        if search.strip() == "":
            return self.__artist
        return search

    def _search_for_artwork(self):
        """
            Search artwork on the web
        """
        ArtworkSearchWidget._search_for_artwork(self)
        self._loaders = 3
        search = self._get_current_search()
        App().task_helper.run(App().artist_art.search_artwork_from_google,
                              search,
                              self._cancellable)
        App().task_helper.run(App().artist_art.search_artwork_from_startpage,
                              search,
                              self._cancellable)
        App().task_helper.run(
                App().artist_art.search,
                search,
                self._cancellable)

    def _on_activate(self, flowbox, child):
        """
            Save artwork
            @param flowbox as Gtk.FlowBox
            @param child as ArtworkSearchChild
        """
        try:
            if isinstance(child, ArtworkSearchChild):
                App().task_helper.run(App().artist_art.add,
                                      self.__artist, child.bytes,
                                      StorageType.COLLECTION)
            else:
                App().task_helper.run(App().artist_art.add,
                                      self.__artist, None,
                                      StorageType.COLLECTION)
            emit_signal(self, "hidden", True)
        except Exception as e:
            Logger.error("ArtistArtworkSearchWidget::_on_activate(): %s", e)

    def __on_artist_artwork(self, surface):
        """
            Set artist artwork
            @param surface as cairo.Surface
        """
        if surface is None:
            self.__artwork.get_style_context().add_class("circle-icon")
            self.__artwork.set_size_request(ArtSize.SMALL, ArtSize.SMALL)
            self.__artwork.set_from_icon_name("avatar-default-symbolic",
                                              Gtk.IconSize.BUTTON)
        else:
            self.__artwork.get_style_context().remove_class("circle-icon")
            self.__artwork.set_from_surface(surface)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from gettext import gettext as _

from lollypop.define import App, Type, LovedFlags
from lollypop.objects_track import Track


class LovedWidget(Gtk.Bin):
    """
        Loved widget
    """

    def __init__(self, object, icon_size=Gtk.IconSize.BUTTON):
        """
            Init widget
            @param object as Album/Track
            @param icon_size as Gtk.IconSize
        """
        Gtk.Bin.__init__(self)
        self.__object = object
        self.__icon_size = icon_size
        self.__timeout_id = None
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/LovedWidget.ui")
        builder.connect_signals(self)
        self.__artwork = builder.get_object("artwork")
        self.add(builder.get_object("widget"))
        self.__set_artwork(self.__object.loved)

    def set_icon_size(self, icon_size):
        """
            Set widget icon size
            @param icon_size as Gtk.IconSize
        """
        self.__icon_size = icon_size
        self.__set_artwork(self.__object.loved)

#######################
# PROTECTED           #
#######################
    def _on_enter_notify_event(self, widget, event):
        """
            Update love opacity
            @param widget as Gtk.EventBox
            @param event as Gdk.Event
        """
        if self.__object.loved & LovedFlags.LOVED:
            loved = LovedFlags.NONE
        else:
            loved = LovedFlags.LOVED
        self.__set_artwork(loved)

    def _on_leave_notify_event(self, widget, event):
        """
            Update love opacity
            @param widget as Gtk.EventBox (can be None)
            @param event as Gdk.Event (can be None)
        """
        self.__set_artwork(self.__object.loved)

    def _on_button_release_event(self, widget, event):
        """
            Toggle loved status
            @param widget as Gtk.EventBox
            @param event as Gdk.Event
        """
        if self.__object.loved & LovedFlags.LOVED:
            loved = LovedFlags.NONE
        else:
            loved = LovedFlags.LOVED
        self.__object.set_loved(loved)
        if isinstance(self.__object, Track):
            # Clear loved playlist artwork cache
            name = App().playlists.get_name(Type.LOVED)
            App().art.remove_from_cache("playlist_" + name, "ROUNDED")
            name = App().playlists.get_name(Type.SKIPPED)
            App().art.remove_from_cache("playlist_" + name, "ROUNDED")
            # Update state on Last.fm
            status = True if loved & LovedFlags.LOVED else False
            for scrobbler in App().ws_director.scrobblers:
                scrobbler.set_loved(self.__object, status)
        self.__set_artwork(self.__object.loved)
        return True

#######################
# PRIVATE             #
#######################
    def __set_artwork(self, flags):
        """
            Set artwork base on object status
            @param flags as int
        """
        if flags & LovedFlags.LOVED:
            self.set_tooltip_text(_("Like"))
            self.__artwork.set_opacity(0.8)
            self.__artwork.set_from_icon_name("emblem-favorite-symbolic",
                                              self.__icon_size)
        else:
            self.set_tooltip_text(_("Allow playback"))
            self.__artwork.set_opacity(0.2)
            self.__artwork.set_from_icon_name("emblem-favorite-symbolic",
                                              self.__icon_size)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk


class GesturesHelper():
    """
        Helper for gesture on widgets
    """

    def __init__(self, widget, primary_long_callback=None,
                 secondary_long_callback=None,
                 primary_press_callback=None,
                 secondary_press_callback=None,
                 tertiary_press_callback=None):
        """
            Init helper
            @param widget as Gtk.Widget
            @params as callbacks
        """
        self.__widget = widget
        self.__allow_multi_released = False
        widget.connect("destroy", self.__on_destroy)
        self.__primary_long_callback = primary_long_callback
        self.__secondary_long_callback = secondary_long_callback
        self.__primary_press_callback = primary_press_callback
        self.__secondary_press_callback = secondary_press_callback
        self.__tertiary_press_callback = tertiary_press_callback
        self.__long_press = Gtk.GestureLongPress.new(widget)
        self.__long_press.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.__long_press.connect("pressed", self.__on_long_pressed)
        self.__long_press.connect("cancelled", self.__on_long_cancelled)
        self.__long_press.set_button(0)
        self.__multi_press = Gtk.GestureMultiPress.new(widget)
        self.__multi_press.connect("released", self.__on_multi_released)
        self.__multi_press.set_propagation_phase(Gtk.PropagationPhase.TARGET)
        self.__multi_press.set_button(0)

    def special_headerbar_hack(self):
        """
            Enable a special header bar hack to block mutter
        """
        self.__multi_press.connect("pressed", self.__on_multi_pressed)

    @property
    def widget(self):
        """
            Get widget
            @return Gtk.Widget
        """
        return self.__widget

    @property
    def multi_press_gesture(self):
        """
            Get multi press gesture
            @return Gtk.GestureMultiPress
        """
        return self.__multi_press

#######################
# PROTECTED           #
#######################
    def _on_primary_long_press_gesture(self, x, y):
        if self.__primary_long_callback is not None:
            self.__primary_long_callback(x, y)

    def _on_secondary_long_press_gesture(self, x, y):
        if self.__secondary_long_callback is not None:
            self.__secondary_long_callback(x, y)

    def _on_primary_press_gesture(self, x, y, event):
        if self.__primary_press_callback is not None:
            self.__primary_press_callback(x, y, event)

    def _on_secondary_press_gesture(self, x, y, event):
        if self.__secondary_press_callback is not None:
            self.__secondary_press_callback(x, y, event)

    def _on_tertiary_press_gesture(self, x, y, event):
        if self.__tertiary_press_callback is not None:
            self.__tertiary_press_callback(x, y, event)

#######################
# PRIVATE             #
#######################
    def __on_destroy(self, widget):
        """
            Remove ref cycle
            @param widget as Gtk.Widget
        """
        self.__primary_long_callback = None
        self.__secondary_long_callback = None
        self.__primary_press_callback = None
        self.__secondary_press_callback = None
        self.__tertiary_press_callback = None
        self.__long_press = None
        self.__multi_press = None
        self.__widget = None

    def __on_long_pressed(self, gesture, x, y):
        """
            Run callback for long pressed
            @param gesture as Gtk.Gesture
            @param x as int
            @param y as int
        """
        if gesture.get_current_button() == 1:
            self._on_primary_long_press_gesture(x, y)
        else:
            self._on_secondary_long_press_gesture(x, y)

    def __on_long_cancelled(self, gesture):
        """
            Allow multi released
            @param gesture as Gtk.Gesture
        """
        self.__allow_multi_released = True

    def __on_multi_pressed(self, gesture, n_press, x, y):
        """
            For headerbar hack
            @param gesture as Gtk.Gesture
            @param n_press as int
            @param x as int
            @param y as int
        """
        if gesture.get_current_button() == 3:
            sequence = gesture.get_current_sequence()
            event = gesture.get_last_event(sequence)
            gesture.set_state(Gtk.EventSequenceState.CLAIMED)
            self._on_secondary_press_gesture(x, y, event)

    def __on_multi_released(self, gesture, n_press, x, y):
        """
            Check released button
            @param gesture as Gtk.Gesture
            @param n_press as int
            @param x as int
            @param y as int
        """
        if not self.__allow_multi_released:
            return
        self.__allow_multi_released = False
        sequence = gesture.get_current_sequence()
        event = gesture.get_last_event(sequence)
        if gesture.get_current_button() == 1:
            self._on_primary_press_gesture(x, y, event)
        elif gesture.get_current_button() == 2:
            self._on_tertiary_press_gesture(x, y, event)
        else:
            self._on_secondary_press_gesture(x, y, event)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


from lollypop.define import App, Type
from lollypop.utils import get_default_storage_type
from lollypop.widgets_albums_rounded import RoundedAlbumsWidget


class AlbumsDecadeWidget(RoundedAlbumsWidget):
    """
        Decade widget showing cover for 4 albums
    """

    def __init__(self, item_ids, view_type, font_height):
        """
            Init widget
            @param decade as [int]
            @param view_type as ViewType
            @param font_height as int
        """
        decade_str = "%s - %s" % (item_ids[0], item_ids[-1])
        RoundedAlbumsWidget.__init__(self, item_ids, decade_str,
                                     decade_str, view_type, font_height)
        self._genre = Type.YEARS

    def populate(self):
        """
            Populate widget content
        """
        if self._artwork is None:
            RoundedAlbumsWidget.populate(self)
        else:
            self.set_artwork()

    @property
    def artwork_name(self):
        """
            Get artwork name
            return str
        """
        return "decade_" + self.name

#######################
# PROTECTED           #
#######################
    def _get_album_ids(self):
        """
            Get album ids
            @return [int]
        """
        storage_type = get_default_storage_type()
        items = []
        for year in self._data:
            items += App().tracks.get_albums_by_disc_for_year(
                                                       year,
                                                       storage_type, True,
                                                       self._ALBUMS_COUNT)
            l = len(items)
            if l < self._ALBUMS_COUNT:
                items += App().tracks.get_compilations_by_disc_for_year(
                                                       year,
                                                       storage_type, True,
                                                       self._ALBUMS_COUNT)
        return [item[0] for item in items]

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango

from gettext import gettext as _

from lollypop.define import App, ArtSize, MARGIN, ViewType
from lollypop.define import SelectionListMask
from lollypop.utils import popup_widget
from lollypop.widgets_banner import BannerWidget
from lollypop.helper_signals import SignalsHelper, signals_map


class PlaylistsBannerWidget(BannerWidget, SignalsHelper):
    """
        Banner for playlists
    """

    @signals_map
    def __init__(self, view):
        """
            Init banner
            @param view as PlaylistView
        """
        BannerWidget.__init__(self, view.args["view_type"] | ViewType.OVERLAY)
        self.__view = view
        grid = Gtk.Grid()
        grid.set_property("valign", Gtk.Align.CENTER)
        grid.get_style_context().add_class("linked")
        grid.show()
        self.__title_label = Gtk.Label.new(
            "<b>" + _("Playlists") + "</b>")
        self.__title_label.show()
        self.__title_label.set_use_markup(True)
        self.__title_label.set_hexpand(True)
        self.__title_label.get_style_context().add_class("dim-label")
        self.__title_label.set_property("halign", Gtk.Align.START)
        self.__title_label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__new_button = Gtk.Button.new_from_icon_name(
            "document-new-symbolic", Gtk.IconSize.BUTTON)
        self.__new_button.connect("clicked", self.__on_new_button_clicked)
        self.__new_button.set_property("halign", Gtk.Align.CENTER)
        self.__new_button.get_style_context().add_class("banner-button")
        self.__new_button.show()
        self.__menu_button = Gtk.Button.new_from_icon_name(
            "view-more-symbolic", Gtk.IconSize.BUTTON)
        self.__menu_button.show()
        self.__menu_button.get_style_context().add_class("banner-button")
        self.__menu_button.set_property("halign", Gtk.Align.END)
        self.__menu_button.connect("clicked", self.__on_menu_button_clicked)
        box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        box.show()
        box.get_style_context().add_class("linked")
        box.add(self.__new_button)
        box.add(self.__menu_button)
        grid.add(self.__title_label)
        grid.add(box)
        grid.set_margin_start(MARGIN)
        grid.set_margin_end(MARGIN)
        self._overlay.add_overlay(grid)
        self._overlay.set_overlay_pass_through(grid, True)
        self.__set_internal_size()
        return [
            (App().window.container.widget, "notify::folded",
             "_on_container_folded"),
        ]

    @property
    def height(self):
        """
            Get wanted height
            @return int
        """
        return ArtSize.SMALL

#######################
# PROTECTED           #
#######################
    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.__set_internal_size()

#######################
# PRIVATE             #
#######################
    def __set_internal_size(self):
        """
            Update font size
        """
        title_context = self.__title_label.get_style_context()
        for c in title_context.list_classes():
            title_context.remove_class(c)
        if App().window.folded:
            self.__title_label.get_style_context().add_class(
                "text-large")
        else:
            self.__title_label.get_style_context().add_class(
                "text-x-large")

    def __on_new_button_clicked(self, button):
        """
            Add a new playlist
            @param button as Gtk.Button
        """
        App().playlists.add(App().playlists.get_new_name())

    def __on_menu_button_clicked(self, button):
        """
            Show playlist menu
            @param button as Gtk.Button
        """
        from lollypop.menu_selectionlist import SelectionListMenu
        from lollypop.widgets_menu import MenuBuilder
        menu = SelectionListMenu(self.__view,
                                 SelectionListMask.PLAYLISTS,
                                 App().window.folded)
        menu_widget = MenuBuilder(menu)
        menu_widget.show()
        popup_widget(menu_widget, button, None, None, button)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk


class ListBox(Gtk.ListBox):
    def __init__(self):
        Gtk.ListBox.__init__(self)
        self.__width = 0

    def set_width(self, width):
        self.set_property("halign", Gtk.Align.CENTER)
        self.__width = width

    def do_get_preferred_width(self):
        if self.__width == 0:
            return Gtk.ListBox.do_get_preferred_width(self)
        else:
            return (self.__width, self.__width)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from random import shuffle, sample
import json

from lollypop.define import LASTFM_API_KEY, App
from lollypop.logger import Logger
from lollypop.utils import emit_signal, get_network_available
from lollypop.helper_web_lastfm import LastFMWebHelper
from lollypop.helper_web_save import SaveWebHelper


class LastFMSimilars(SaveWebHelper, LastFMWebHelper):
    """
        Search similar artists with Last.fm
    """
    def __init__(self):
        """
            Init provider
        """
        SaveWebHelper.__init__(self)
        LastFMWebHelper.__init__(self)

    def load_similars(self, artist_ids, storage_type, cancellable):
        """
            Load similar artists for artist ids
            @param artist_ids as int
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
        """
        names = [App().artists.get_name(artist_id) for artist_id in artist_ids]
        result = self.get_similar_artists(names, cancellable)
        tracks = []
        for (artist_name, cover_uri) in result:
            albums = self.get_artist_top_albums(artist_name, cancellable)
            albums = sample(albums, min(len(albums), 10))
            if not albums:
                continue
            for (album, artist) in albums:
                payload = self.get_album_payload(album, artist, cancellable)
                if payload is None:
                    continue
                lollypop_payload = self.lollypop_album_payload(payload)
                item = self.save_album_payload_to_db(lollypop_payload,
                                                     storage_type,
                                                     True,
                                                     cancellable)
                tracks = sample(payload["tracks"]["track"],
                                min(len(payload["tracks"]["track"]), 3))
                shuffle(tracks)
                i = 1
                for track in tracks:
                    lollypop_payload = self.lollypop_track_payload(track, i)
                    i += 1
                    self.save_track_payload_to_db(lollypop_payload,
                                                  item,
                                                  storage_type,
                                                  True,
                                                  cancellable)
        emit_signal(self, "finished")

    def get_similar_artists(self, artist_names, cancellable):
        """
            Search similar artists
            @param artist_names as [str]
            @param cancellable as Gio.Cancellable
            @return [(str, str)] as [(artist_name, cover_uri)]
        """
        result = []
        for artist_name in artist_names:
            try:
                for similar in self.__get_similar_artists(artist_name):
                    if cancellable.is_cancelled():
                        raise Exception("cancelled")
                    result.append((similar, None))
            except Exception as e:
                Logger.error("LastFMSimilars::get_similar_artists(): %s", e)
        if result:
            Logger.info("Found similar artists with LastFMSimilars")
        return result

#######################
# PRIVATE             #
#######################
    def __get_similar_artists(self, artist):
        """
            Get similar artists
            @param artist as str
            @return similars as [str]
        """
        if not get_network_available("LASTFM"):
            return []
        artists = []
        try:
            uri = "http://ws.audioscrobbler.com/2.0/?method=artist.getinfo"
            uri += "&artist=%s&api_key=%s&format=json" % (
                artist, LASTFM_API_KEY)
            (status, data) = App().task_helper.load_uri_content_sync(uri, None)
            if status:
                content = json.loads(data.decode("utf-8"))
                for artist in content["artist"]["similar"]["artist"]:
                    artists.append(artist["name"])
        except:
            Logger.error(
                "LastFMWebHelper::__get_similar_artists(): %s", data)
        return artists



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from gettext import gettext as _

from lollypop.define import App
from lollypop.utils import get_network_available


class SearchMenu(Gio.Menu):
    """
        Contextual menu for search
    """
    def __init__(self, header=False):
        """
            Init search menu
            @param header as bool
        """
        Gio.Menu.__init__(self)
        if header:
            from lollypop.menu_header import MenuHeader
            self.append_item(MenuHeader(_("Search"), "edit-find-symbolic"))
        search_action = Gio.SimpleAction.new_stateful(
            "web_search",
            GLib.VariantType.new("s"),
            GLib.Variant("s", "NONE"))
        App().add_action(search_action)
        section = Gio.Menu()
        self.append_section(_("Search on the Web"), section)
        menu_item = Gio.MenuItem.new(_("Disabled"),
                                     "app.web_search('NONE')")
        section.append_item(menu_item)

        if not get_network_available("YOUTUBE"):
            return

        web_search = App().settings.get_value("web-search")
        if web_search.get_string() != "NONE":
            if get_network_available(web_search.get_string()):
                search_action.set_state(web_search)
        search_action.connect("change-state",
                              self.__on_search_change_state)

        if get_network_available("DEEZER"):
            menu_item = Gio.MenuItem.new(_("Deezer"),
                                         "app.web_search('DEEZER')")
            section.append_item(menu_item)
        if get_network_available("LASTFM"):
            menu_item = Gio.MenuItem.new(_("Last.fm"),
                                         "app.web_search('LASTFM')")
            section.append_item(menu_item)
        if get_network_available("MUSICBRAINZ"):
            menu_item = Gio.MenuItem.new(_("MusicBrainz"),
                                         "app.web_search('MUSICBRAINZ')")
            section.append_item(menu_item)
        if get_network_available("SPOTIFY"):
            menu_item = Gio.MenuItem.new(_("Spotify"),
                                         "app.web_search('SPOTIFY')")
            section.append_item(menu_item)
        if get_network_available("JAMENDO"):
            menu_item = Gio.MenuItem.new(_("Jamendo"),
                                         "app.web_search('JAMENDO')")
            section.append_item(menu_item)

#######################
# PRIVATE             #
#######################
    def __on_search_change_state(self, action, value):
        """
            Update search setting
            @param action as Gio.SimpleAction
            @param value as bool
        """
        App().settings.set_value("web-search", value)
        action.set_state(value)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Pango

from gettext import gettext as _

from lollypop.view import View
from lollypop.define import App, ViewType
from lollypop.define import StorageType, LYRICS_PATH
from lollypop.logger import Logger
from lollypop.utils import get_network_available
from lollypop.objects_track import Track
from lollypop.helper_lyrics import LyricsHelper
from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.widgets_banner_lyrics import LyricsBannerWidget
from lollypop.information_store import InformationStore


class LyricsLabel(Gtk.Stack):
    """
        Lyrics label with crossfade on change
    """

    def __init__(self):
        """
            Init label
        """
        Gtk.Stack.__init__(self)
        self.__label1 = Gtk.Label.new()
        self.__label1.set_line_wrap_mode(Pango.WrapMode.WORD)
        self.__label1.set_line_wrap(True)
        self.__label1.set_justify(Gtk.Justification.CENTER)
        self.__label2 = Gtk.Label.new()
        self.__label2.set_line_wrap_mode(Pango.WrapMode.WORD)
        self.__label2.set_line_wrap(True)
        self.__label2.set_justify(Gtk.Justification.CENTER)
        self.__label1.show()
        self.__label2.show()
        self.add(self.__label1)
        self.add(self.__label2)
        self.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.set_transition_duration(250)

    def set_text(self, text, crossfade=False):
        """
            Set label text
            @param text as str
            @param crossfade as bool
        """
        if crossfade:
            self.next()
        self.get_visible_child().set_text(text)

    def set_markup(self, markup, crossfade=False):
        """
            Set label markup
            @param markup as str
            @param crossfade as bool
        """
        if self.get_visible_child().get_label() == markup:
            return
        if crossfade:
            self.next()
        self.get_visible_child().set_markup(markup)

    def next(self):
        """
            Show next label
        """
        for child in self.get_children():
            if child != self.get_visible_child():
                self.set_visible_child(child)
                break


class LyricsView(View, SignalsHelper):
    """
        Show lyrics for track
    """

    __FILTERS = ["[explicit]", "(explicit)"]

    # TODO add https://www.musixmatch.com support
    @signals_map
    def __init__(self, view_type):
        """
            Init view
            @param view_type as ViewType
        """
        View.__init__(self, StorageType.COLLECTION,
                      view_type | ViewType.OVERLAY)
        self.__lyrics_timeout_id = None
        self.__downloads_running = 0
        self.__lyrics_text = ""
        self._empty_message = _("No track playing")
        self._empty_icon_name = "view-dual-symbolic"
        self.__lyrics_label = LyricsLabel()
        self.__lyrics_label.show()
        self.__lyrics_label.set_property("halign", Gtk.Align.CENTER)
        self.__banner = LyricsBannerWidget(self.view_type)
        self.__banner.show()
        self.__banner.connect("translate", self.__on_translate)
        self.add_widget(self.__lyrics_label, self.__banner)
        self.__lyrics_helper = LyricsHelper()
        self.__update_lyrics_style()
        self.__information_store = InformationStore()
        return [
            (App().window.container.widget, "notify::folded",
             "_on_container_folded"),
            (App().player, "current-changed", "_on_current_changed")
        ]

    def populate(self, track):
        """
            Set lyrics
            @param track as Track
        """
        self.__banner.translate_button.set_sensitive(False)
        if track.id is None:
            self.__lyrics_label.set_text("")
            return
        self.__lyrics_label.set_text(_("Loading…"))
        lyrics = ""
        if isinstance(track, Track):
            self.__lyrics_helper.load(track)
            # First check synced lyrics
            if self.__lyrics_helper.available:
                if self.__lyrics_timeout_id is None:
                    self.__lyrics_timeout_id = GLib.timeout_add(
                        200, self.__show_sync_lyrics)
                return
            else:
                if self.__lyrics_timeout_id is not None:
                    GLib.source_remove(self.__lyrics_timeout_id)
                    self.__lyrics_timeout_id = None
                if track.storage_type & (StorageType.COLLECTION |
                                         StorageType.EXTERNAL):
                    from lollypop.tagreader import TagReader, Discoverer
                    tagreader = TagReader()
                    discoverer = Discoverer()
                    try:
                        info = discoverer.get_info(track.uri)
                    except:
                        info = None
                    if info is not None:
                        tags = info.get_tags()
                        lyrics = tagreader.get_lyrics(tags)
        if lyrics:
            self.__lyrics_label.set_text(lyrics)
            self.__lyrics_text = lyrics
        else:
            name = track.name + track.album.name + ",".join(track.artists)
            content = self.__information_store.get_information(name,
                                                               LYRICS_PATH)
            if content:
                self.__lyrics_label.set_text(content.decode("utf-8"))
            elif not get_network_available():
                self.__lyrics_label.set_text(
                    _("Network unavailable or disabled in settings"))
            else:
                self.__lyrics_helper.get_lyrics_from_web(track,
                                                         self.__on_lyrics,
                                                         False,
                                                         track)

    @property
    def args(self):
        return None

##############
# PROTECTED  #
##############
    def __on_translate(self, banner, active):
        """
            Translate lyrics
            @param banner as LyricsBannerWidget
            @param active as bool
        """
        if active:
            App().task_helper.run(self.__get_blob, self.__lyrics_text,
                                  callback=(self.__lyrics_label.set_text,))
        else:
            self.__lyrics_label.set_text(self.__lyrics_text)

    def _on_unmap(self, widget):
        """
            Connect player signal
            @param widget as Gtk.Widget
        """
        self.__lyrics_helper.cancel()
        View._on_unmap(self, widget)
        if self.__lyrics_timeout_id is not None:
            GLib.source_remove(self.__lyrics_timeout_id)
            self.__lyrics_timeout_id = None

    def _on_current_changed(self, player):
        """
            Update lyrics
            @param player as Player
        """
        self.populate(App().player.current_track)

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.__update_lyrics_style()

############
# PRIVATE  #
############
    def __show_sync_lyrics(self):
        """
            Show sync lyrics for track
        """
        timestamp = App().player.position
        (previous, current, next) =\
            self.__lyrics_helper.get_lyrics_for_timestamp(timestamp + 200)
        lyrics = ""
        if previous:
            opacity = 20
            delta = 50 // len(previous)
            for line in previous:
                if line.strip():
                    escaped = GLib.markup_escape_text(line)
                    lyrics += "<span alpha='{}%'>{}\n</span>".format(opacity,
                                                                     escaped)
                    opacity += delta
        else:
            opacity = 80
        for line in current:
            if line.strip():
                escaped = GLib.markup_escape_text(line)
                lyrics += "<span size='x-large'>%s\n</span>" % escaped
        if next:
            delta = opacity // len(next)
            for line in next:
                if line.strip():
                    escaped = GLib.markup_escape_text(line)
                    lyrics += "<span alpha='{}%'>{}\n</span>".format(opacity,
                                                                     escaped)
                    opacity -= delta
        self.__lyrics_label.set_markup(lyrics, True)
        return True

    def __get_blob(self, text):
        """
            Translate text with current user locale
            @param text as str
        """
        try:
            locales = GLib.get_language_names()
            user_code = locales[0].split(".")[0]
            try:
                from textblob.blob import TextBlob
            except:
                return _("You need to install python3-textblob module")
            blob = TextBlob(text)
            return str(blob.translate(to=user_code))
        except Exception as e:
            Logger.error("LyricsView::__get_blob(): %s", e)
            return _("Can't translate this lyrics")

    def __update_lyrics_style(self):
        """
            Update lyrics style based on current view width
        """
        context = self.get_style_context()
        for cls in context.list_classes():
            context.remove_class(cls)
        context.add_class("lyrics")
        if App().window.folded:
            if self.__lyrics_helper.available:
                context.add_class("text-large")
            else:
                context.add_class("text-medium")
        else:
            if self.__lyrics_helper.available:
                context.add_class("text-x-large")
            else:
                context.add_class("text-large")

    def __on_lyrics(self, lyrics, filtered, track):
        """
            Set lyrics
            @param lyrics as str/None
            @param filtered as bool
            @param track as Track
        """
        if lyrics is None:
            self.__lyrics_label.set_text(
                    _("Disabled in network settings"))
        elif lyrics == "":
            if filtered:
                self.__lyrics_label.set_text(_("No lyrics found ") + "😓")
            else:
                name = track.name.lower()
                track = Track(track.id)
                for _filter in self.__FILTERS:
                    name = name.replace(_filter, "")
                track.set_name(name)
                self.__lyrics_helper.get_lyrics_from_web(track,
                                                         self.__on_lyrics,
                                                         True,
                                                         track)
        else:
            self.__lyrics_label.set_text(lyrics)
            self.__lyrics_text = lyrics
            name = track.name + track.album.name + ",".join(track.artists)
            self.__information_store.save_information(name,
                                                      LYRICS_PATH,
                                                      lyrics.encode("utf-8"))
            self.__banner.translate_button.set_sensitive(True)



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

import json

from lollypop.logger import Logger
from lollypop.utils import emit_signal, get_network_available
from lollypop.helper_web_lastfm import LastFMWebHelper
from lollypop.helper_web_save import SaveWebHelper
from lollypop.define import LASTFM_API_KEY, App


class LastFMSearch(LastFMWebHelper, SaveWebHelper):
    """
        Search for LastFM
    """

    def __init__(self):
        """
            Init object
        """
        SaveWebHelper.__init__(self)
        LastFMWebHelper.__init__(self)

    def get(self, search, storage_type, cancellable):
        """
            Get albums for search
            We need a thread because we are going to populate DB
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
        """
        if not get_network_available("LASTFM"):
            emit_signal(self, "finished")
            return
        try:
            uri = "http://ws.audioscrobbler.com/2.0/?method=album.search"
            uri += "&album=%s&api_key=%s&format=json" % (
                search, LASTFM_API_KEY)
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            albums = []
            if status:
                decode = json.loads(data.decode("utf-8"))
                for album in decode["results"]["albummatches"]["album"]:
                    albums.append((album["name"], album["artist"]))
            for (album, artist) in albums:
                uri = "http://ws.audioscrobbler.com/2.0/?method=album.getinfo"
                uri += "&api_key=%s&artist=%s&album=%s&format=json" % (
                    LASTFM_API_KEY,
                    GLib.uri_escape_string(artist, None, True),
                    GLib.uri_escape_string(album, None, True))
                (status, data) = App().task_helper.load_uri_content_sync(
                    uri, cancellable)
                if status:
                    decode = json.loads(data.decode("utf-8"))
                    try:
                        payload = self.lollypop_album_payload(decode["album"])
                        item = self.save_album_payload_to_db(payload,
                                                             storage_type,
                                                             True,
                                                             cancellable)
                        i = 1
                        for track in decode["album"]["tracks"]["track"]:
                            payload = self.lollypop_track_payload(track, i)
                            i += 1
                            self.save_track_payload_to_db(payload,
                                                          item,
                                                          storage_type,
                                                          True,
                                                          cancellable)
                    except Exception as e:
                        Logger.warning("LastFMSearch::get(): %s", e)
        except Exception as e:
            Logger.warning("LastFMSearch::get(): %s", e)
        if not cancellable.is_cancelled():
            emit_signal(self, "finished")

    def load_tracks(self, album_id, storage_type, cancellable):
        pass



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

from random import choice
from gettext import gettext as _

from lollypop.logger import Logger
from lollypop.objects_album import Album
from lollypop.player_auto_similar import AutoSimilarPlayer
from lollypop.player_auto_random import AutoRandomPlayer
from lollypop.define import App, Repeat
from lollypop.utils import emit_signal


class AlbumsPlayer:
    """
        Handle player albums
    """

    def __init__(self):
        """
            Init player
        """
        # Albums in current playlist
        self._albums = []

    def add_album(self, album):
        """
            Add album to player
            @param album as Album
        """
        self.add_albums([album])

    def add_album_ids(self, album_ids):
        """
            Add album ids to player
            @param album_ids as [int]
        """
        self.add_albums([Album(album_id) for album_id in album_ids])

    def add_albums(self, albums):
        """
            Add albums to player
            @param albums as [Album]
        """
        if not albums:
            App().notify.send(_("No album available"))
            return
        try:
            for album in albums:
                # Merge album if previous is same
                if self._albums and self._albums[-1].id == album.id:
                    track_ids = self._albums[-1].track_ids
                    for track in album.tracks:
                        if track.id not in track_ids:
                            self._albums[-1].append_track(track)
                    emit_signal(self, "playback-updated", self._albums[-1])
                else:
                    self._albums.append(album)
                    emit_signal(self, "playback-added", album)
            self.update_next_prev()
        except Exception as e:
            Logger.error("Player::add_albums(): %s" % e)

    def remove_album(self, album):
        """
            Remove album from albums
            @param album as Album
        """
        try:
            if album not in self._albums:
                return
            self._albums.remove(album)
            emit_signal(self, "playback-removed", album)
            self.update_next_prev()
        except Exception as e:
            Logger.error("Player::remove_album(): %s" % e)

    def remove_album_by_id(self, album_id):
        """
            Remove all instances of album id
            @param album_id as int
        """
        self.remove_album_by_ids([album_id])

    def remove_album_by_ids(self, album_ids):
        """
            Remove all instances of album ids
            @param album_ids as [int]
        """
        try:
            for album_id in album_ids:
                for album in self._albums:
                    if album.id == album_id:
                        self.remove_album(album)
                        emit_signal(self, "playback-removed", album)
            self.update_next_prev()
        except Exception as e:
            Logger.error("Player::remove_album_by_ids(): %s" % e)

    def remove_track_from_album(self, track, album):
        """
            Remove track from album
            @param track as Track
            @param album as Album
        """
        is_current_track = track == self.current_track
        if is_current_track:
            self.next()
        if album.remove_track(track):
            self.remove_album(album)
            emit_signal(self, "playback-removed", album)
        else:
            emit_signal(self, "playback-updated", album)
            if not is_current_track:
                self.update_next_prev()

    def play_album(self, album):
        """
            Play album
            @param album as Album
        """
        self.play_album_for_albums(album, [album])

    def play_albums(self, albums):
        """
            Play albums
            @param album as [Album]
        """
        if not albums:
            App().notify.send(_("No album available"))
            return
        if App().settings.get_value("shuffle"):
            album = choice(albums)
        else:
            album = albums[0]
        self.play_album_for_albums(album, albums)

    def play_track_for_albums(self, track, albums):
        """
            Play track and set albums as current playlist
            @param albums as [Album]
            @param track as Track
        """
        if self.is_party:
            App().lookup_action("party").change_state(GLib.Variant("b", False))
        self._albums = albums
        self.load(track)
        emit_signal(self, "playback-setted", list(albums))

    def play_album_for_albums(self, album, albums):
        """
            Play album and set albums as current playlist
            @param album as Album
            @param albums as [Album]
        """
        if not albums:
            App().notify.send(_("No album available"))
            return
        if self.is_party:
            App().lookup_action("party").change_state(GLib.Variant("b", False))
        if App().settings.get_value("shuffle"):
            self.__play_shuffle_tracks(album, albums)
        else:
            self.__play_albums(album, albums)

    def set_albums(self, albums, signal=True):
        """
            Set player albums
            @param albums as [Album]
            @param signal as bool
        """
        if not albums:
            App().notify.send(_("No album available"))
            return
        self._albums = albums
        if signal:
            emit_signal(self, "playback-setted", list(albums))
        self.update_next_prev()

    def clear_albums(self):
        """
            Clear all albums
        """
        self._albums = []
        emit_signal(self, "playback-setted", [])
        self.update_next_prev()

    def skip_album(self):
        """
            Skip current album
        """
        try:
            # In party or shuffle, just update next track
            if self.is_party or App().settings.get_value("shuffle"):
                self.set_next()
            elif self._current_track.id is not None:
                index = self._albums.index(
                    self._current_track.album)
                if index + 1 >= len(self._albums):
                    repeat = App().settings.get_enum("repeat")
                    if repeat == Repeat.AUTO_SIMILAR:
                        next_album = AutoSimilarPlayer.next_album(self)
                        if next_album is not None:
                            self.add_album(next_album)
                    elif repeat == Repeat.AUTO_RANDOM:
                        next_album = AutoRandomPlayer.next_album(self)
                        if next_album is not None:
                            self.add_album(next_album)
                    elif repeat == Repeat.ALL:
                        next_album = self._albums[0]
                    else:
                        next_album = None
                else:
                    next_album = self._albums[index + 1]
                if next_album is None:
                    self.stop()
                else:
                    self.load(next_album.tracks[0])
        except Exception as e:
            Logger.error("Player::skip_album(): %s" % e)

    def track_in_playback(self, track):
        """
            True if track present in current playback
            @param track as Track
            @return Track/None
        """
        for album in self._albums:
            if album.id == track.album.id:
                for _track in album.tracks:
                    if track.id == _track.id:
                        return _track
        return None

    def get_albums_for_id(self, album_id):
        """
            Get albums for id
            @param album_id as int
            @return [Album]
        """
        return [album for album in self._albums if album.id == album_id]

    @property
    def albums(self):
        """
            Return albums
            @return albums as [Album]
        """
        return list(self._albums)

    @property
    def album_ids(self):
        """
            Return albums ids
            @return albums ids as [int]
        """
        return [album.id for album in self._albums]

#######################
# PRIVATE             #
#######################
    def __play_shuffle_tracks(self, album, albums):
        """
            Start shuffle tracks playback.
            @param album as Album
            @param albums as [albums]
        """
        if album is None:
            album = choice(albums)
        if album.tracks:
            track = choice(album.tracks)
        else:
            track = None
        self._albums = albums
        emit_signal(self, "playback-setted", list(albums))
        if track is not None:
            self.load(track)
        else:
            self.update_next_prev()

    def __play_albums(self, album, albums):
        """
            Start albums playback.
            @param album as Album
            @param albums as [albums]
        """
        if album is None:
            album = albums[0]
        if album.tracks:
            track = album.tracks[0]
        else:
            track = None
        self._albums = albums
        emit_signal(self, "playback-setted", list(albums))
        if track is not None:
            self.load(track)
        else:
            self.update_next_prev()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from gettext import gettext as _

from lollypop.widgets_row_track import TrackRow
from lollypop.widgets_listbox import ListBox
from lollypop.define import App, ViewType, Size, MARGIN
from lollypop.objects_track import Track
from lollypop.helper_signals import SignalsHelper, signals
from lollypop.helper_gestures import GesturesHelper


class QueueTracksView(Gtk.Bin, GesturesHelper, SignalsHelper):
    """
        Responsive view showing queued tracks
    """

    @signals
    def __init__(self):
        """
            Init view
        """
        Gtk.Bin.__init__(self)
        label = Gtk.Label.new(_("Currently in queue"))
        label.show()
        label.get_style_context().add_class("dim-label")
        label.get_style_context().add_class("large")
        self.__box = ListBox()
        self.__box.show()
        self.__box.set_selection_mode(Gtk.SelectionMode.NONE)
        self.__box.set_width(Size.SMALL)
        self.__box.connect("row-activated", self.__on_row_activated)
        self.__grid = Gtk.Grid()
        self.__grid.set_orientation(Gtk.Orientation.VERTICAL)
        self.__grid.set_halign(Gtk.Align.CENTER)
        self.__grid.add(label)
        self.__grid.add(self.__box)
        self.__grid.set_margin_bottom(MARGIN)
        self.__grid.get_style_context().add_class("borders")
        self.add(self.__grid)
        GesturesHelper.__init__(self, self.__box)
        return [
            (App().player, "queue-changed", "_on_queue_changed"),
            (App().player, "current-changed", "_on_current_changed"),
        ]

    def populate(self):
        """
            Populate with current queue
        """
        self.allow_duplicate("_on_queue_changed")
        tracks = [Track(track_id) for track_id in App().player.queue]
        self.__add_tracks(tracks)

#######################
# PROTECTED           #
#######################
    def _on_queue_changed(self, *ignore):
        """
            Clean view and reload if empty
        """
        queue = App().player.queue
        track_ids = [row.track.id for row in self.__box.get_children()]
        for track_id in queue:
            if track_id not in track_ids:
                self.__add_tracks([Track(track_id)])
        for row in self.__box.get_children():
            if row.track.id not in queue:
                row.destroy()
        if not self.__box.get_children():
            self.__grid.hide()

    def _on_current_changed(self, player):
        """
            Update children state
            @param player as Player
        """
        for child in self.__box.get_children():
            child.update_number_label()

    def _on_secondary_press_gesture(self, x, y, event):
        """
            Show row menu
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        row = self.__box.get_row_at_y(y)
        if row is None:
            return
        row.popup_menu(self.__box, x, y)

#######################
# PRIVATE             #
#######################
    def __add_tracks(self, tracks):
        """
            Add tracks to widget
            @param tracks as [Track]
        """
        for track in tracks:
            row = TrackRow(track, [], ViewType.QUEUE)
            row.show()
            row.connect("removed", self.__on_track_removed)
            self.__box.add(row)
            self.__grid.show()

    def __on_row_activated(self, listbox, row):
        """
            Play track
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        App().player.load(row.track)

    def __on_track_removed(self, row):
        """
            Remove track from queue
            @param row as TrackRow
        """
        if row.track.id in App().player.queue:
            App().player.remove_from_queue(row.track.id)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk


from lollypop.widgets_playlist_smart import SmartPlaylistRow
from lollypop.view import View
from lollypop.logger import Logger
from lollypop.define import App, StorageType


class SmartPlaylistView(View):
    """
        Show a view allowing user to create a smart playlist
    """

    def __init__(self, playlist_id, view_type):
        """
            Init PlaylistView
            @param playlist_id as int
            @param view_type as ViewType
        """
        View.__init__(self, StorageType.ALL, view_type)
        self.__playlist_id = playlist_id
        self.__size_group = Gtk.SizeGroup()
        self.__size_group.set_mode(Gtk.SizeGroupMode.BOTH)
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/SmartPlaylistView.ui")
        widget = builder.get_object("widget")
        self.connect("size-allocate", self.__on_size_allocate, widget)
        self.__listbox = builder.get_object("listbox")
        self.scrolled.set_property("expand", True)
        self.__match_toggle = builder.get_object("match_toggle")
        self.__operand_combobox = builder.get_object("operand_combobox")
        self.__select_combobox = builder.get_object("select_combobox")
        self.__limit_spin = builder.get_object("limit_spin")
        self.__add_rule_button = builder.get_object("add_rule_button")
        self.__up_box = builder.get_object("up_box")
        self.__bottom_box = builder.get_object("bottom_box")
        if App().playlists.get_smart(playlist_id):
            self.__match_toggle.set_active(True)
            self.__set_sensitive(True)
        self.add_widget(widget)
        builder.connect_signals(self)

    def populate(self):
        """
            Setup an initial widget based on current request
        """
        sql = App().playlists.get_smart_sql(self.__playlist_id)
        if sql is None:
            return
        try:
            if sql.find(" UNION ") != -1:
                operand = "OR"
            else:
                operand = "AND"
            self.__operand_combobox.set_active_id(operand)
        except Exception as e:
            self.__operand_combobox.set_active(0)
            Logger.warning("SmartPlaylistView::populate: %s", e)
        # Setup rows
        for line in sql.split("((")[1:]:
            widget = SmartPlaylistRow(self.__size_group)
            try:
                widget.set(line.split("))")[0])
            except Exception as e:
                Logger.error("SmartPlaylistView::populate: %s", e)
            widget.show()
            self.__listbox.add(widget)
        try:
            split_limit = sql.split("LIMIT")
            limit = int(split_limit[1].split(" ")[1])
            self.__limit_spin.set_value(limit)
        except Exception as e:
            Logger.warning("SmartPlaylistView::populate: %s", e)
        try:
            split_order = sql.split("ORDER BY")
            split_spaces = split_order[1].split(" ")
            orderby = split_spaces[1]
            if split_spaces[2] in ["ASC", "DESC"]:
                orderby += " %s" % split_spaces[2]
            self.__select_combobox.set_active_id(orderby)
        except Exception as e:
            self.__select_combobox.set_active(0)
            Logger.warning("SmartPlaylistView::populate: %s", e)

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"playlist_id": self.__playlist_id,
                "view_type": self.view_type}

#######################
# PROTECTED           #
#######################
    def __get_or_request(self):
        """
            Get request for AND operand
            @return str
        """
        request = ""
        orderby = self.__select_combobox.get_active_id()
        for child in self.__listbox.get_children():
            if child.sql is None:
                continue
            request += "SELECT DISTINCT(tracks.rowid)"
            if orderby != "random()":
                request += ", %s" % orderby
            request += " FROM tracks"
            subrequest = " WHERE %s" % child.sql
            if orderby.find("albums.name") != -1:
                subrequest += " AND tracks.album_id = albums.rowid"
            elif orderby.find("artists.name") != -1:
                subrequest += " AND track_artists.artist_id = artists.rowid\
                                AND track_artists.track_id = tracks.rowid"
            if subrequest.find(" genres."):
                request += ", %s" % "genres"
            if subrequest.find(" artists."):
                request += ", %s" % "artists"
            if subrequest.find(" albums.") != -1:
                request += ", %s" % "albums"
            if subrequest.find(" album_genres.") != -1:
                request += ", %s" % " album_genres"
            if subrequest.find(" track_artists.") != -1:
                request += ", %s" % " track_artists"
            request += subrequest + " UNION "
        request = request[:-7]  # " UNION "
        request += " ORDER BY %s" % orderby
        request += " LIMIT %s" % int(self.__limit_spin.get_value())
        return request

    def __get_and_request(self):
        """
            Get request for AND operand
            @return str
        """
        orderby = self.__select_combobox.get_active_id()
        request = "SELECT DISTINCT(tracks.rowid) FROM tracks"
        subrequest = " WHERE"
        for child in self.__listbox.get_children():
            if child.sql is not None:
                subrequest += " %s AND" % child.sql
        subrequest = subrequest[:-3]
        if orderby.find("albums.name") != -1:
            subrequest += " AND tracks.album_id = albums.rowid"
        elif orderby.find("artists.name") != -1:
            subrequest += " AND track_artists.artist_id = artists.rowid\
                            AND track_artists.track_id = tracks.rowid"
        if subrequest.find(" genres.") != -1:
            request += ", %s" % "genres"
        if subrequest.find(" artists.") != -1:
            request += ", %s" % "artists"
        if subrequest.find(" albums.") != -1:
            request += ", %s" % "albums"
        if subrequest.find(" album_genres.") != -1:
            request += ", %s" % "album_genres"
        if subrequest.find(" track_artists.") != -1:
            request += ", %s" % "track_artists"
        subrequest += " ORDER BY %s" % orderby
        subrequest += " LIMIT %s" % int(self.__limit_spin.get_value())
        return request + subrequest

    def _on_save_button_clicked(self, button):
        """
            Save SQL request
            @param button as Gtk.Button
        """
        operand = self.__operand_combobox.get_active_id()
        if len(self.__listbox.get_children()) == 0:
            request = ""
            App().playlists.set_smart(self.__playlist_id, False)
        elif operand == "AND":
            request = self.__get_and_request()
        else:
            request = self.__get_or_request()
        App().playlists.set_smart_sql(self.__playlist_id, request)
        App().window.container.go_back()

    def _on_add_rule_button_clicked(self, button):
        """
            Add a new rule
            @param button as Gtk.Button
        """
        self.__populate()

    def _on_match_check_button_toggled(self, button):
        """
            Enable/Disable smart playlist
            @param button as GtkButton
        """
        active = button.get_active()
        App().playlists.set_smart(self.__playlist_id, active)
        self.__set_sensitive(active)

    def _on_map(self, widget):
        """
            Disable global shortcuts and update toolbar
            @param widget as Gtk.Widget
        """
        View._on_map(self, widget)
        App().enable_special_shortcuts(False)

    def _on_unmap(self, widget):
        """
            Enable global shortcuts
            @param widget as Gtk.Widget
        """
        View._on_unmap(self, widget)
        App().enable_special_shortcuts(True)

#######################
# PRIVATE             #
#######################
    def __populate(self):
        """
            Setup an initial widget
        """
        widget = SmartPlaylistRow(self.__size_group)
        widget.show()
        self.__listbox.add(widget)

    def __set_sensitive(self, sensitive):
        """
            Set view sensitive
            @param sensitive as bool
        """
        if sensitive:
            self.__up_box.set_sensitive(True)
            self.__bottom_box.set_sensitive(True)
            self.__add_rule_button.set_sensitive(True)
            self.__listbox.set_sensitive(True)
        else:
            self.__up_box.set_sensitive(False)
            self.__bottom_box.set_sensitive(False)
            self.__add_rule_button.set_sensitive(False)
            self.__listbox.set_sensitive(False)

    def __on_size_allocate(self, widget, allocation, child_widget):
        """
            Set child widget size
            @param widget as Gtk.Widget
            @param allocation as Gtk.Allocation
            @param child_widget as Gtk.Widget
        """
        width = max(400, allocation.width / 2)
        child_widget.set_size_request(width, -1)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, GLib, GObject, Pango

from locale import strcoll

from lollypop.view_lazyloading import LazyLoadingView
from lollypop.helper_gestures import GesturesHelper
from lollypop.fastscroll import FastScroll
from lollypop.define import Type, App, ArtSize, SelectionListMask
from lollypop.define import ArtBehaviour, ViewType, StorageType
from lollypop.logger import Logger
from lollypop.utils import get_icon_name, on_query_tooltip, popup_widget
from lollypop.utils import emit_signal


class SelectionListRow(Gtk.ListBoxRow):
    """
        A selection list row
    """

    __gsignals__ = {
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def get_best_height(widget):
        """
            Calculate widget height
            @param widget as Gtk.Widget
        """
        ctx = widget.get_pango_context()
        layout = Pango.Layout.new(ctx)
        layout.set_text("a", 1)
        font_height = int(layout.get_pixel_size()[1])
        return font_height

    def __init__(self, rowid, name, sortname, mask, height):
        """
            Init row
            @param rowid as int
            @param name as str
            @param sortname as str
            @param mask as SelectionListMask
            @param height as str
        """
        Gtk.ListBoxRow.__init__(self)
        self.__artwork = None
        self.__rowid = rowid
        self.__name = name
        self.__sortname = sortname
        self.__mask = mask
        self.set_style(height)

    def populate(self):
        """
            Populate widget
        """
        if self.__rowid == Type.SEPARATOR:
            separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
            separator.show()
            self.add(separator)
            self.set_sensitive(False)
            emit_signal(self, "populated")
        else:
            self.__grid = Gtk.Grid()
            self.__grid.set_column_spacing(7)
            self.__grid.show()
            self.__artwork = Gtk.Image.new()
            self.__grid.add(self.__artwork)
            self.__label = Gtk.Label.new()
            self.__label.set_markup(GLib.markup_escape_text(self.__name))
            self.__label.set_property("has-tooltip", True)
            self.__label.connect("query-tooltip", on_query_tooltip)
            self.__label.set_xalign(0)
            self.__grid.add(self.__label)
            if self.__mask & SelectionListMask.ARTISTS:
                self.__grid.set_margin_end(20)
            self.add(self.__grid)
            self.set_artwork()
            self.set_mask()

    def set_label(self, string):
        """
            Set label for row
            @param string as str
        """
        self.__name = string
        if not self.__mask & SelectionListMask.SIDEBAR:
            self.__label.set_markup(GLib.markup_escape_text(string))

    def set_artwork(self):
        """
            set_artwork widget
        """
        if self.__artwork is None:
            return
        if self.__rowid == Type.SEPARATOR:
            pass
        elif self.__mask & SelectionListMask.ARTISTS and\
                self.__rowid >= 0 and\
                App().settings.get_value("artist-artwork"):
            App().art_helper.set_artist_artwork(
                                    self.__name,
                                    ArtSize.SMALL,
                                    ArtSize.SMALL,
                                    self.get_scale_factor(),
                                    ArtBehaviour.ROUNDED |
                                    ArtBehaviour.CROP_SQUARE |
                                    ArtBehaviour.CACHE,
                                    self.__on_artist_artwork)
            self.__artwork.show()
        elif self.__rowid < 0:
            icon_name = get_icon_name(self.__rowid)
            self.__artwork.set_from_icon_name(icon_name, Gtk.IconSize.INVALID)
            self.__artwork.set_pixel_size(20)
            self.__artwork.show()
            emit_signal(self, "populated")
        else:
            self.__artwork.hide()
            emit_signal(self, "populated")

    def set_mask(self, mask=None):
        """
            Set row mask
            @param mask as SelectionListMask
        """
        # Do nothing if widget not populated
        if self.__artwork is None:
            self.__mask = mask
            return
        # Do not update widget if mask does not changed
        elif mask == self.__mask:
            return
        # If no mask, use current one
        elif mask is None:
            mask = self.__mask
        # Else use new mask
        else:
            self.__mask = mask

        if mask & (SelectionListMask.LABEL |
                   SelectionListMask.ARTISTS |
                   SelectionListMask.GENRES):
            self.__artwork.set_property("halign", Gtk.Align.FILL)
            self.__artwork.set_hexpand(False)
            self.__label.show()
            self.set_tooltip_text("")
            self.set_has_tooltip(False)
        else:
            self.__artwork.set_property("halign", Gtk.Align.CENTER)
            self.__artwork.set_hexpand(True)
            self.__label.hide()
            self.set_tooltip_text(self.__label.get_text())
            self.set_has_tooltip(True)
        if mask & SelectionListMask.ELLIPSIZE:
            self.__label.set_ellipsize(Pango.EllipsizeMode.END)
        else:
            self.__label.set_ellipsize(Pango.EllipsizeMode.NONE)

    def set_style(self, height):
        """
            Set internal sizing
        """
        if self.__rowid == Type.SEPARATOR:
            height = -1
            self.set_sensitive(False)
        elif self.__mask & SelectionListMask.ARTISTS and\
                self.__rowid >= 0 and\
                App().settings.get_value("artist-artwork"):
            self.get_style_context().add_class("row")
            if height < ArtSize.SMALL:
                height = ArtSize.SMALL
            # Padding => application.css
            height += 12
        elif self.__mask & SelectionListMask.SIDEBAR:
            self.get_style_context().add_class("row-big")
            # Padding => application.css
            height += 30
        else:
            self.get_style_context().add_class("row")
        self.set_size_request(-1, height)

    @property
    def is_populated(self):
        """
            Return True if populated
            @return bool
        """
        return self.get_child() is not None

    @property
    def name(self):
        """
            Get row name
            @return str
        """
        return self.__name

    @property
    def sortname(self):
        """
            Get row sortname
            @return str
        """
        return self.__sortname

    @property
    def id(self):
        """
            Get row id
            @return int
        """
        return self.__rowid

#######################
# PRIVATE             #
#######################
    def __on_artist_artwork(self, surface):
        """
            Set artist artwork
            @param surface as cairo.Surface
        """
        if surface is None:
            self.__artwork.get_style_context().add_class("circle-icon")
            self.__artwork.set_size_request(ArtSize.SMALL,
                                            ArtSize.SMALL)
            self.__artwork.set_from_icon_name(
                                              "avatar-default-symbolic",
                                              Gtk.IconSize.DND)
        else:
            self.__artwork.get_style_context().remove_class("circle-icon")
            self.__artwork.set_from_surface(surface)
        emit_signal(self, "populated")


class SelectionList(LazyLoadingView, GesturesHelper):
    """
        A list for artists/genres
    """
    __gsignals__ = {
        "expanded": (GObject.SignalFlags.RUN_FIRST, None, (bool,))
    }

    def __init__(self, base_mask):
        """
            Init Selection list ui
            @param base_mask as SelectionListMask
        """
        LazyLoadingView.__init__(self, StorageType.ALL, ViewType.DEFAULT)
        self.__selection_pending_ids = []
        self.__base_mask = base_mask
        self.__mask = SelectionListMask.NONE
        self.__animation_timeout_id = None
        self.__height = SelectionListRow.get_best_height(self)
        self._box = Gtk.ListBox()
        self._box.set_selection_mode(Gtk.SelectionMode.MULTIPLE)
        self._box.show()
        GesturesHelper.__init__(self, self._box)
        self.__scrolled = Gtk.ScrolledWindow()
        self.__scrolled.show()
        self.__scrolled.set_policy(Gtk.PolicyType.NEVER,
                                   Gtk.PolicyType.AUTOMATIC)
        self.__viewport = Gtk.Viewport()
        self.__scrolled.add(self.__viewport)
        self.__viewport.show()
        self.__viewport.add(self._box)
        self.connect("initialized", self.__on_initialized)
        self.get_style_context().add_class("sidebar")
        self.__scrolled.set_vexpand(True)
        if base_mask & SelectionListMask.FASTSCROLL:
            self.__overlay = Gtk.Overlay.new()
            self.__overlay.show()
            self.__overlay.add(self.__scrolled)
            self.__fastscroll = FastScroll(self._box,
                                           self.__scrolled)
            self.__overlay.add_overlay(self.__fastscroll)
            self.add(self.__overlay)
            App().settings.connect("changed::artist-artwork",
                                   self.__on_artist_artwork_changed)
            App().artist_art.connect("artist-artwork-changed",
                                     self.__on_artist_artwork_changed)
        else:
            self.__overlay = None
            App().settings.connect("changed::show-sidebar-labels",
                                   self.__on_show_sidebar_labels_changed)
            self.add(self.__scrolled)
            self.__menu_button = Gtk.Button.new_from_icon_name(
                "view-more-horizontal-symbolic", Gtk.IconSize.BUTTON)
            self.__menu_button.set_property("halign", Gtk.Align.CENTER)
            self.__menu_button.get_style_context().add_class("no-border")
            self.__menu_button.connect("clicked",
                                       lambda x: self.__popup_menu(None, x))
            self.__menu_button.show()
            self.add(self.__menu_button)
        if base_mask & SelectionListMask.SIDEBAR:
            App().window.container.widget.connect("notify::folded",
                                                  self.__on_container_folded)
            self.__on_container_folded(None, App().window.folded)
        else:
            self.__base_mask |= SelectionListMask.LABEL
            self.__base_mask |= SelectionListMask.ELLIPSIZE
            self.__set_rows_mask(self.__base_mask | self.__mask)

    def set_mask(self, mask):
        """
            Mark list as artists list
            @param mask as SelectionListMask
        """
        self.__mask = mask

    def add_mask(self, mask):
        """
            Mark list as artists list
            @param mask as SelectionListMask
        """
        self.__mask |= mask

    def populate(self, values):
        """
            Populate view with values
            @param [(int, str, optional str)], will be deleted
        """
        self._box.set_sort_func(None)
        self.__scrolled.get_vadjustment().set_value(0)
        self.clear()
        LazyLoadingView.populate(self, values)

    def remove_value(self, object_id):
        """
            Remove id from list
            @param object_id as int
        """
        for child in self._box.get_children():
            if child.id == object_id:
                child.destroy()
                break

    def add_value(self, value):
        """
            Add item to list
            @param value as (int, str, optional str)
        """
        self._box.set_sort_func(self.__sort_func)
        ids = [row.id for row in self._box.get_children()]
        if value[0] not in ids:
            child = self._get_child(value)
            child.populate()
            if self.mask & SelectionListMask.ARTISTS:
                self.__fastscroll.clear()
                self.__fastscroll.populate()

    def update_value(self, object_id, name):
        """
            Update object with new name
            @param object_id as int
            @param name as str
        """
        found = False
        for child in self._box.get_children():
            if child.id == object_id:
                child.set_label(name)
                found = True
                break
        if not found:
            if self.__base_mask & SelectionListMask.FASTSCROLL:
                self.__fastscroll.clear()
            self.add_value((object_id, name, name))

    def update_values(self, values):
        """
            Update view with values
            @param [(int, str, optional str)]
        """
        if self.mask & SelectionListMask.FASTSCROLL:
            self.__fastscroll.clear()
        # Remove not found items
        value_ids = set([v[0] for v in values])
        for child in self._box.get_children():
            if child.id not in value_ids:
                self.remove_value(child.id)
        # Add items which are not already in the list
        item_ids = set([child.id for child in self._box.get_children()])
        for value in values:
            if not value[0] in item_ids:
                row = self._get_child(value)
                row.populate()
        if self.mask & SelectionListMask.ARTISTS:
            self.__fastscroll.populate()

    def select_ids(self, ids=[], activate=True):
        """
            Select listbox items
            @param ids as [int]
            @param activate as bool
        """
        if ids:
            rows = []
            for row in self._box.get_children():
                if row.id in ids:
                    rows.append(row)
            if rows:
                self._box.unselect_all()
                for row in rows:
                    self._box.select_row(row)
                if activate:
                    rows[0].activate()
        else:
            self._box.unselect_all()

    def clear(self):
        """
            Clear treeview
        """
        self.stop()
        for child in self._box.get_children():
            child.destroy()
        if self.__base_mask & SelectionListMask.FASTSCROLL:
            self.__fastscroll.clear()
            self.__fastscroll.clear_chars()

    def select_first(self):
        """
            Select first available item
        """
        try:
            self._box.unselect_all()
            row = self._box.get_children()[0]
            self._box.select_row(row)
            row.activate()
        except Exception as e:
            Logger.warning("SelectionList::select_first(): %s", e)

    def set_selection_pending_ids(self, pending_ids):
        """
            Set selection pending ids
            @param pending_ids
        """
        self.__selection_pending_ids = pending_ids

    def select_pending_ids(self):
        """
            Select pending ids
        """
        self.select_ids(self.__selection_pending_ids)
        self.__selection_pending_ids = []

    def activate_child(self):
        """
            Activated typeahead row
        """
        self._box.unselect_all()
        for row in self._box.get_children():
            style_context = row.get_style_context()
            if style_context.has_class("typeahead"):
                row.activate()
            style_context.remove_class("typeahead")

    @property
    def filtered(self):
        """
            Get filtered children
            @return [Gtk.Widget]
        """
        filtered = []
        for child in self._box.get_children():
            if isinstance(child, SelectionListRow):
                filtered.append(child)
        return filtered

    @property
    def overlay(self):
        """
            Get list overlay
            @return overlay as Gtk.Overlay
        """
        return self.__overlay

    @property
    def listbox(self):
        """
            Get listbox
            @return Gtk.ListBox
        """
        return self._box

    @property
    def mask(self):
        """
            Get selection list type
            @return bit mask
        """
        return self.__mask | self.__base_mask

    @property
    def args(self):
        return None

    @property
    def count(self):
        """
            Get items count in list
            @return int
        """
        return len(self._box.get_children())

    @property
    def selected_ids(self):
        """
            Get selected ids
            @return [int]
        """
        return [row.id for row in self._box.get_selected_rows()]

    @property
    def scrolled(self):
        """
            Get scrolled window
            @return Gtk.ScrolledWindow
        """
        return self.__scrolled

    @property
    def selected_id(self):
        """
            Get selected id
            @return int
        """
        selected_row = self._box.get_selected_row()
        return None if selected_row is None else selected_row.id

#######################
# PROTECTED           #
#######################
    def _get_child(self, value):
        """
            Get a child for view
            @param value as [(int, str, optional str)]
            @return row as SelectionListRow
        """
        (rowid, name, sortname) = value
        if rowid > 0 and self.mask & SelectionListMask.ARTISTS:
            used = sortname if sortname else name
            self.__fastscroll.add_char(used[0])
        row = SelectionListRow(rowid, name, sortname,
                               self.mask, self.__height)
        row.show()
        self._box.add(row)
        return row

    def _scroll_to_child(self, row):
        """
            Scroll to row
            @param row as SelectionListRow
        """
        coordinates = row.translate_coordinates(self._box, 0, 0)
        if coordinates:
            self.__scrolled.get_vadjustment().set_value(coordinates[1])

    def _on_primary_long_press_gesture(self, x, y):
        """
            Show row menu
            @param x as int
            @param y as int
        """
        self.__popup_menu(y)

    def _on_primary_press_gesture(self, x, y, event):
        """
            Activate current row
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        row = self._box.get_row_at_y(y)
        if row is not None:
            (exists, state) = event.get_state()
            if state & Gdk.ModifierType.CONTROL_MASK or\
                    state & Gdk.ModifierType.SHIFT_MASK:
                pass
            else:
                self._box.unselect_all()

    def _on_secondary_press_gesture(self, x, y, event):
        """
            Show row menu
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        self.__popup_menu(y)

#######################
# PRIVATE             #
#######################
    def __set_rows_mask(self, mask):
        """
            Show labels on child
            @param status as bool
        """
        for row in self._box.get_children():
            row.set_mask(mask)

    def __sort_func(self, row_a, row_b):
        """
            Sort rows
            @param row_a as SelectionListRow
            @param row_b as SelectionListRow
        """
        a_index = row_a.id
        b_index = row_b.id

        # Static vs static
        if a_index < 0 and b_index < 0:
            return a_index < b_index
        # Static entries always on top
        elif b_index < 0:
            return True
        # Static entries always on top
        if a_index < 0:
            return False
        # String comparaison for non static
        else:
            if self.mask & SelectionListMask.ARTISTS:
                a = row_a.sortname
                b = row_b.sortname
            else:
                a = row_a.name
                b = row_b.name
            return strcoll(a, b)

    def __popup_menu(self, y=None, relative=None):
        """
            Show menu at y or row
            @param y as int
            @param relative as Gtk.Widget
        """
        if self.__base_mask & SelectionListMask.SIDEBAR:
            menu = None
            row_id = None
            if relative is None:
                relative = self._box.get_row_at_y(y)
                if relative is not None:
                    row_id = relative.id
            if row_id is None:
                from lollypop.menu_selectionlist import SelectionListMenu
                menu = SelectionListMenu(self,
                                         self.mask,
                                         App().window.folded)
            elif not App().settings.get_value("save-state"):
                from lollypop.menu_selectionlist import SelectionListRowMenu
                menu = SelectionListRowMenu(row_id,
                                            App().window.folded)
            if menu is not None:
                from lollypop.widgets_menu import MenuBuilder
                menu_widget = MenuBuilder(menu)
                menu_widget.show()
                popup_widget(menu_widget, relative, None, None, None)

    def __on_artist_artwork_changed(self, object, value):
        """
            Update row artwork
            @param object as GObject.Object
            @param value as str
        """
        artist = value if object == App().art else None
        if self.mask & SelectionListMask.ARTISTS:
            for row in self._box.get_children():
                if artist is None:
                    row.set_style(self.__height)
                    row.set_artwork()
                elif row.name == artist:
                    row.set_artwork()
                    break

    def __on_show_sidebar_labels_changed(self, settings, value):
        """
            Show/hide labels
            @param settings as Gio.Settings
            @param value as str
        """
        self.__on_container_folded(None, App().window.folded)

    def __on_initialized(self, selectionlist):
        """
            Update fastscroll
            @param selectionlist as SelectionList
        """
        if self.mask & SelectionListMask.ARTISTS:
            self.__fastscroll.populate()
        # Scroll to first selected item
        for row in self._box.get_selected_rows():
            GLib.idle_add(self._scroll_to_child, row)
            break

    def __on_container_folded(self, leaflet, folded):
        """
            Update internals
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        folded = App().window.folded
        self.__base_mask &= ~SelectionListMask.LABEL
        self.__scrolled.set_hexpand(folded)
        if folded or App().settings.get_value("show-sidebar-labels"):
            self.__base_mask |= SelectionListMask.LABEL
        self.__set_rows_mask(self.__base_mask | self.__mask)
